#!/usr/bin/env python
#
# Splits stream of XML elements into etree docs.
#
# Author: Just van den Broecke
#
from stetl.util import Util, etree
from stetl.filter import Filter
from stetl.packet import FORMAT

log = Util.get_log('xmlassembler')


class XmlAssembler(Filter):
    """
    Split a stream of etree DOM XML elements (usually Features) into etree DOM docs.
    Consumes and buffers elements until max_elements reached, will then produce an etree doc.

    consumes=FORMAT.etree_element, produces=FORMAT.etree_doc
    """
    xpath_base = "//*[local-name() = '%s']"

    # Constructor
    def __init__(self, configdict, section):
        Filter.__init__(self, configdict, section, consumes=FORMAT.etree_element, produces=FORMAT.etree_doc)

        log.info("cfg = %s" % self.cfg.to_string())
        self.max_elements = self.cfg.get_int('max_elements', 10000)
        self.container_doc = self.cfg.get('container_doc')
        self.element_container_xpath = XmlAssembler.xpath_base % self.cfg.get('element_container_tag')
        self.total_element_count = 0
        self.element_arr = []

        # Reusable XML parser
        self.xml_parser = etree.XMLParser(remove_blank_text=True)

    def invoke(self, packet):
        if packet.data is not None:
            # Valid element: consume and handle
            self.consume_element(packet)

            # Document is obviously not finished, reset EoD/EoS in packet
            packet.set_end_of_stream(False)
            packet.set_end_of_doc(False)

        if packet.is_end_of_stream() or packet.is_end_of_doc() or len(self.element_arr) >= self.max_elements:
            # EOF but still data in buffer: make doc
            # log.info("Flush doc")
            self.flush_elements(packet)

        return packet

    def consume_element(self, packet):
        # Always move the data (element) from packet
        element = packet.consume()

        if element is not None:
            self.total_element_count += 1
            self.element_arr.append(element)

        return packet

    def flush_elements(self, packet):
        packet.set_end_of_doc()

        if len(self.element_arr) == 0:
            return packet

        # Start new doc (TODO clone)
        try:
            etree_doc = etree.fromstring(self.container_doc.encode('utf-8'), self.xml_parser)
        except Exception as e:
            log.error('new container doc not OK: %s' % str(e))
            return packet

        parent_element = etree_doc.xpath(self.element_container_xpath)
        if len(parent_element) > 0:
            parent_element = parent_element[0]

        for element in self.element_arr:
            parent_element.append(element)

        log.info('xmldoc ready: elms=%d total_elms=%d' % (len(self.element_arr), self.total_element_count))
        packet.data = etree_doc
        self.element_arr = []
        return packet



__import__('pkg_resources').declare_namespace(__name__)

# from chain import *
# from component import Config
# from etl import *
# from factory import *
# from filter import *
# from input import *
# from output import *
# from packet import *
# from postgis import *
# from util import *



import os
import sys

from stetl.etl import ETL
from stetl.inputs.fileinput import LineStreamerFileInput
from stetl.filters.nullfilter import NullFilter
from stetl.outputs.standardoutput import StandardOutput
from stetl.merger import Merger
from tests.stetl_test_case import StetlTestCase


class MergerMultiInputTest(StetlTestCase):
    """Unit tests for Merger"""
    pass

    def setUp(self):
        super(MergerMultiInputTest, self).setUp()

        # Initialize Stetl
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(curr_dir, 'configs/mergermultiinput.cfg')}
        self.etl = ETL(cfg_dict)

    def test_instance(self):
        # Chain #1 - Simple Case
        chain = StetlTestCase.get_chain(self.etl)

        merger_comp = chain.first_comp
        self.assertTrue(isinstance(merger_comp, Merger))
        self.assertEqual(len(merger_comp.children), 2)
        self.assertTrue(isinstance(merger_comp.children[0][0], LineStreamerFileInput),
                        "Next is not LineStreamerFileInput")
        self.assertTrue(isinstance(merger_comp.children[1][0], LineStreamerFileInput),
                        "Next is not LineStreamerFileInput")
        self.assertTrue(isinstance(merger_comp.children[0][1], LineStreamerFileInput),
                        "Next is not LineStreamerFileInput")
        self.assertTrue(isinstance(merger_comp.children[1][1], LineStreamerFileInput),
                        "Next is not LineStreamerFileInput")
        self.assertTrue(isinstance(merger_comp.children[0][1].next, StandardOutput),
                        "Next is not StandardOutput")
        self.assertTrue(isinstance(merger_comp.children[1][1].next, StandardOutput),
                        "Next is not StandardOutput")

        # Flag for End-of-Stream 2 subcomps
        self.assertEqual(merger_comp.end_count, 2)

        # Chain #2 - SubChain Case
        chain = StetlTestCase.get_chain(self.etl, index=1)

        merger_comp = chain.first_comp
        children = merger_comp.children
        self.assertTrue(isinstance(merger_comp, Merger))
        self.assertEqual(len(merger_comp.children), 2, "Child count is not 2")
        self.assertTrue(isinstance(merger_comp.first(children[0]), LineStreamerFileInput),
                        "Next is not LineStreamerFileInput")
        self.assertTrue(isinstance(merger_comp.first(children[1]), LineStreamerFileInput),
                        "Next is not LineStreamerFileInput")
        self.assertTrue(isinstance(merger_comp.children[0][0].next, NullFilter),
                        "Next is not NullFilter")
        self.assertTrue(isinstance(merger_comp.first(children[1]).next, NullFilter),
                        "Next is not NullFilter")
        self.assertTrue(isinstance(merger_comp.last(children[1]).next, StandardOutput),
                        "Next is not StandardOutput")
        self.assertTrue(isinstance(merger_comp.children[1][1].next, StandardOutput),
                        "Next is not StandardOutput")

        # Flag for End-of-Stream 2 subcomps
        self.assertEqual(merger_comp.end_count, 2)

    def test_execute(self):
        run_count = 0
        for index in [0, 1]:
            chain = StetlTestCase.get_chain(self.etl, index=index)
            chain.run()

            merger_comp = chain.first_comp

            # Flag for End-of-Stream 2 subcomps
            self.assertEqual(merger_comp.end_count, 0)

            # Result should be merged lines from both files
            result = sys.stdout.getvalue().split('\n')

            # Strip empty lines
            result = [s for s in result if (s or len(s) > 0)]

            # Total should be sum of linecount non-empty lines in input files
            # each run increases line count hence multiply by run_count
            run_count += 1
            self.assertEqual(len(result), 28 * run_count)



#!/usr/bin/env python
#
# Splits stream of GML lines into etree docs.
#
# Author: Just van den Broecke
#
import codecs
from deprecated.sphinx import deprecated
from stetl.util import Util, etree, StringIO
from stetl.filter import Filter
from stetl.packet import FORMAT

log = Util.get_log('gmlsplitter')


@deprecated(version='1.0.4', reason='Use the more robust XmlElementStreamerFileInput + XmlAssembler instead!!!')
class GmlSplitter(Filter):
    """
    Split a stream of text XML lines into documents
    TODO phase out

    consumes=FORMAT.xml_line_stream, produces=FORMAT.etree_doc
    """

    def __init__(self, configdict, section='gml_splitter'):
        Filter.__init__(self, configdict, section, consumes=FORMAT.xml_line_stream, produces=FORMAT.etree_doc)

        log.info("cfg = %s" % self.cfg.to_string())
        self.max_features = self.cfg.get_int('max_features', 10000)
        # File preamble
        self.start_container = self.cfg.get('start_container')
        # File postamble
        self.end_container = self.cfg.get('end_container')
        self.container_tag = self.cfg.get('container_tag')
        #        self.feature_tags = self.cfg.get('feature_tags').split(',')
        self.start_feature_markers = self.cfg.get('start_feature_markers').split(',')
        self.end_feature_markers = self.cfg.get('end_feature_markers').split(',')
        self.feature_count = 0
        self.total_feature_count = 0
        self.in_heading = True

        # End of file is line with end_container_tag
        self.end_container_tag = '</%s' % self.container_tag

        # Derive start and end tags from feature_tags
        #        for feature_tag in self.feature_tags:
        #            self.start_feature_markers.append('<%s' % feature_tag)
        #            self.end_feature_markers.append('</%s>' % feature_tag)
        self.expect_end_feature_markers = []
        self.expect_end_feature_tag = None
        self.buffer = None
        # Reusable XML parser
        self.xml_parser = etree.XMLParser(remove_blank_text=True)

    def invoke(self, packet):
        if packet.data is None:
            if (packet.is_end_of_stream() or packet.is_end_of_doc()) and self.buffer:
                # EOF but still data in buffer: make doc
                log.info("Last Buffer")
                self.buffer_to_doc(packet)
        else:
            # Valid line: push to the splitter
            # If buffer filled process it
            self.push_line(packet)
            if packet.is_end_of_doc() is True:
                self.buffer_to_doc(packet)

        return packet

    def push_line(self, packet):
        # Assume GML lines with single tag per line !!
        # This is ok for e.g. OGR output
        # TODO we need to become more sophisticated when no newlines in XML

        line = packet.consume()

        if packet.is_end_of_stream() is True:
            return packet

        # Start new buffer filling
        if self.buffer is None:
            self.buffer = self.init_buf()
            self.feature_count = 0
            self.in_heading = True
            packet.set_end_of_doc(False)

        if self.is_start_feature(line) is True:
            if self.in_heading:
                # First time: we are in heading
                # Write container start
                self.buffer.write(self.start_container)
                self.in_heading = False

            self.buffer.write(line)
            self.buffer.write('<!-- Feature #%s -->\n' % self.feature_count)

        else:
            # If within feature or end-of-feature: write
            if self.expect_end_feature_tag is not None and packet.is_end_of_doc() is False:
                self.buffer.write(line)

            # If endtag of feature found may also indicate buffer filled with max_features
            if self.is_end_feature(line) is True:
                # Start or end tag of ogr:feature  element
                if self.feature_count >= self.max_features and self.expect_end_feature_tag is None:
                    self.buffer.write(self.end_container)
                    log.info("Buffer filled feat_count = %d total_feat_count= %d" % (
                        self.feature_count, self.total_feature_count))
                    self.feature_count = 0
                    packet.set_end_of_doc()
                    return packet

            # Last tag (end of container) reaching
            if line.find(self.end_container_tag) >= 0:
                if self.buffer is not None and self.feature_count > 0:
                    self.buffer.write(self.end_container)
                    log.info("Buffer filled (EOF) feat_count = %d total_feat_count= %d" % (
                        self.feature_count, self.total_feature_count))
                    self.feature_count = 0
                    packet.set_end_of_doc()
                    return packet

        return packet

    def buffer_to_doc(self, packet):
        # Process/transform data in buffer
        self.buffer.seek(0)
        try:
            packet.data = etree.parse(self.buffer, self.xml_parser)
        except Exception as e:
            bufStr = self.buffer.getvalue()
            if not bufStr:
                log.info("parse buffer empty: content=[%s]" % bufStr)
            else:
                log.error("error in buffer parsing %s" % str(e))
                print(bufStr)
                raise
        self.buffer.close()
        self.buffer = None

    def init_buf(self):
        buffer = StringIO()
        buffer = codecs.getwriter("utf8")(buffer)
        return buffer

    def is_start_feature(self, line):
        index = 0
        for feature_tag in self.start_feature_markers:
            if line.find(feature_tag) >= 0:
                # found it ! Now we expect the end tag for this start_tag
                self.expect_end_feature_tag = self.end_feature_markers[index]
                self.expect_end_feature_markers.append(self.expect_end_feature_tag)
                self.feature_count += 1
                self.total_feature_count += 1
                return True

            # Not found: next tag
            index += 1

        # Nothing found
        return False

    def is_end_feature(self, line):
        # Not even within a feature ?
        if self.expect_end_feature_tag is None:
            return False

        # ASSERTION: one or more expect_end_feature_tag set, thus within feature

        # End-of-feature reached ?
        if line.find(self.expect_end_feature_tag) >= 0:
            self.expect_end_feature_tag = None
            self.expect_end_feature_markers.pop()
            if len(self.expect_end_feature_markers) > 0:
                # Set expected end-tag to last in list
                self.expect_end_feature_tag = self.expect_end_feature_markers[-1]

            return True

        # Still within feature
        return False



import os

from stetl.etl import ETL
from stetl.filters.xmlassembler import XmlAssembler
from stetl.filters.packetbuffer import PacketBuffer
from tests.stetl_test_case import StetlTestCase

class XmlAssemblerTest(StetlTestCase):
    """Unit tests for XmlAssembler"""

    def setUp(self):
        super(XmlAssemblerTest, self).setUp()

        # Initialize Stetl
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(curr_dir, 'configs/xmlassembler.cfg')}
        self.etl = ETL(cfg_dict)
    
    def test_class(self):
        chain = StetlTestCase.get_chain(self.etl)
        section = StetlTestCase.get_section(chain, 1)
        class_name = self.etl.configdict.get(section, 'class')
        
        self.assertEqual('stetl.filters.xmlassembler.XmlAssembler', class_name)
    
    def test_instance(self):
        chain = StetlTestCase.get_chain(self.etl)

        self.assertTrue(isinstance(chain.get_by_index(1), XmlAssembler))
        
    def test_execute(self):
        chain = StetlTestCase.get_chain(self.etl)
        chain.run()

        buffer_filter = chain.get_by_class(PacketBuffer)
        packet_list = buffer_filter.packet_list

        # most Packets are empty, but we need to find 2 filled with etree docs
        doc_packet_list = []
        for packet in packet_list:
            if packet.data is not None:
                doc_packet_list.append(packet)

        # Assertion: we need to see 2 documents
        self.assertEqual(len(doc_packet_list), 4)
        namespaces={'gml': 'http://www.opengis.net/gml/3.2', 'top10nl': 'http://register.geostandaarden.nl/gmlapplicatieschema/top10nl/1.2.0'}

        # Assertion: first doc has two FeatureMember elements with proper Namespaces
        xml_doc1 = doc_packet_list[0].data
        feature_elms = xml_doc1.xpath('/gml:FeatureCollectionT10NL/top10nl:FeatureMember', namespaces=namespaces)
        self.assertEqual(len(feature_elms), 2)

        # Assertion: last doc has one FeatureMember with proper Namespaces
        last = len(doc_packet_list) - 1
        xml_doc2 = doc_packet_list[last].data
        feature_elms = xml_doc2.xpath('/gml:FeatureCollectionT10NL/top10nl:FeatureMember', namespaces=namespaces)
        self.assertEqual(len(feature_elms), 1)

        # Assertion: first doc has end_of_doc but not end_of_stream set
        self.assertTrue(doc_packet_list[0].end_of_doc, msg='doc1: end_of_doc if False')
        self.assertFalse(doc_packet_list[0].end_of_stream, msg='doc1: end_of_stream is True')

        # Assertion: last doc has end_of_doc and end_of_stream set
        self.assertTrue(doc_packet_list[last].end_of_doc, msg='doc2: end_of_doc if False')
        self.assertTrue(doc_packet_list[last].end_of_stream, msg='doc2: end_of_stream if False')



# Input classes for fetching data via HTTP.
#
# Author: Just van den Broecke
#
import re
from urllib.parse import urlencode
from urllib.request import Request, urlopen
from urllib.error import URLError, HTTPError
import base64

from stetl.component import Config
from stetl.input import Input
from stetl.util import Util
from stetl.packet import FORMAT

log = Util.get_log('httpinput')


class HttpInput(Input):
    """
    Fetch data from remote services like WFS via HTTP protocol.
    Base class: subclasses will do datatype-specific formatting of
    the returned data.

    produces=FORMAT.any
    """

    # Start attribute config meta
    # Applying Decorator pattern with the Config class to provide
    # read-only config values from the configured properties.

    @Config(ptype=str, default=None, required=True)
    def url(self):
        """
        The HTTP URL string.
        """
        pass

    @Config(ptype=dict, default=None, required=False)
    def auth(self):
        """
        Authentication data: Flat JSON-like struct  dependent on auth type/schema.
        Only the `type` field is required, other fields depend on auth schema.
        Supported values : ::

            type: basic|token

        If the type is ``basic`` (HTTP Basic Authentication) two additional fields ``user``
        and ``password`` are required.
        If the type is ``token`` (HTTP Token) additional two additional fields ``keyword``
        and ``token`` are required.

        Any required Base64 encoding is provided by ``HttpInput``.

        Examples: ::

            # Basic Auth
            url = https://some.rest.api.com
            auth = {
                type: basic,
                user: myname
                password: mypassword
            }

            # Token Auth
            url = https://some.rest.api.com
            auth = {
                type: token,
                keyword: Bearer
                token: mytoken
            }
        """
        pass

    @Config(ptype=dict, default=None, required=False)
    def parameters(self):
        """
        Flat JSON-like struct of the parameters to be appended to the url.

        Example: (parameters require quotes)::

            url = http://geodata.nationaalgeoregister.nl/natura2000/wfs
            parameters = {
                service : WFS,
                version : 1.1.0,
                request : GetFeature,
                srsName : EPSG:28992,
                outputFormat : text/xml; subtype=gml/2.1.2,
                typename : natura2000
            }
        """
        pass

    # End attribute config meta

    def __init__(self, configdict, section, produces=FORMAT.any):
        Input.__init__(self, configdict, section, produces)

        log.info("url=%s parameters=%s" % (self.url, self.parameters))

    def add_authorization(self, request):
        """
        Add authorization from config data. Authorization scheme-specific.
        May be extended or overloaded for additional schemes.

        :param request: the HTTP Request
        :return:
        """
        auth_creds = self.auth
        auth_type = auth_creds['type']
        auth_val = None
        if auth_type == 'basic':
            # Basic auth: http://mozgovipc.blogspot.nl/2012/06/python-http-basic-authentication-with.html
            # base64 encode username and password
            # write the Authorization header like: 'Basic base64encode(username + ':' + password)
            auth_val = base64.encodestring('%s:%s' % (auth_creds['user'], auth_creds['password']))
            auth_val = "Basic %s" % auth_val
        elif auth_type == 'token':
            # Bearer Type, see eg. https://tools.ietf.org/html/rfc6750
            auth_val = "%s %s" % (auth_creds['keyword'], auth_creds['token'])

        request.add_header("Authorization", auth_val.replace('\n', ''))

    def read_from_url(self, url, parameters=None):
        """
        Read the data from the URL.

        :param url: the url to fetch
        :param parameters: optional dict of query parameters
        :return:
        """
        # log.info('Fetch data from URL: %s ...' % url)

        req = Request(url)
        try:
            # Urlencode optional parameters
            query_string = None
            if parameters:
                query_string = str.encode(urlencode(parameters))

            # Add optional Authorization
            if self.auth:
                self.add_authorization(req)

            response = urlopen(req, query_string)
        except HTTPError as e:
            log.error('HTTPError fetching from URL %s: code=%d e=%s' % (url, e.code, e))
            raise e
        except URLError as e:
            log.error('URLError fetching from URL %s: reason=%s e=%s' % (url, e.reason, e))
            raise e

        # Everything is fine
        return response.read()

    def read(self, packet):
        """
        Read the data from the URL.

        :param packet:
        :return:
        """
        # Done with URL ?
        if self.url is None:
            packet.set_end_of_stream()
            log.info("EOF URL reading done")
            return packet

        packet.data = self.format_data(self.read_from_url(self.url, self.parameters))
        self.url = None
        return packet

    def format_data(self, data):
        """
        Format response data, override in subclasses, defaults to returning original data.
        :param packet:
        :return:
        """
        return data


class ApacheDirInput(HttpInput):
    """
     Read file data from an Apache directory "index" HTML page.
     Uses http://stackoverflow.com/questions/686147/url-tree-walker-in-python
     produces=FORMAT.record. Each record contains file_name and file_data (other meta data like
     date time is too fragile over different Apache servers).
    """

    @Config(ptype=str, default='xml', required=False)
    def file_ext(self):
        """
        The file extension for target files in Apache dir.
        """
        pass

    def __init__(self, configdict, section, produces=FORMAT.record):
        HttpInput.__init__(self, configdict, section, produces)
        # look for a link + a timestamp + a size ('-' for dir)
        # self.parse_re = re.compile('href="([^"]*)".*(..-...-.... ..:..).*?(\d+[^\s<]*|-)')
        # This appeared to be too fragile, e.g. different date formats per apache server

        # default file extension to filter
        # default regular expression for file
        self.file_reg_exp = self.cfg.get('file_reg_exp', 'href="([^"]*%s)"' % self.file_ext)
        self.parse_re = re.compile(self.file_reg_exp)
        self.file_list = None
        self.file_index = None

        if not self.url.endswith('/'):
            self.url += '/'

    def init(self):
        """
        Read the list of files from the Apache index URL.
        """
        # One time: get all files from remote Apache dir
        log.info('Init: fetching file list from URL: %s ...' % self.url)
        html = self.read_from_url(self.url)
        self.file_list = self.parse_re.findall(html)
        log.info('Found %4d file' % len(self.file_list) + 's' * (len(self.file_list) != 1))

    def next_file(self):
        """
        Return a tuple (name, date, size) with next file info.

        :return tuple:
        """

        if self.file_index is None:
            self.file_index = -1

        # At last file tuple ?
        if self.no_more_files():
            return None

        self.file_index += 1

        return self.file_list[self.file_index]

    def no_more_files(self):
        """
        More files left?.

        :return Boolean:
        """
        return self.file_index == len(self.file_list) - 1

    def read(self, packet):
        """
        Read the data from the URL.

        :param packet:
        :return:
        """
        file_name = self.next_file()

        file_name = self.filter_file(file_name)

        # All files done?
        if file_name is None and self.no_more_files() is True:
            packet.set_end_of_stream()
            log.info("EOF Apache dir files done, file_index=%d" % self.file_index)
            return packet

        if file_name is None:
            return packet

        # Process next file
        url = self.url + file_name
        log.info("Reading file_index=%d, file_name=%s " % (self.file_index, file_name))

        # Create record from file_name and file content
        packet.data = dict(file_name=file_name, file_data=self.read_from_url(url))

        return packet

    def filter_file(self, file_name):
        """
        Filter the file_name, e.g. to suppress reading, default: return file_name.

        :param file_name:
        :return string or None:
        """
        return file_name



# testing: to be called by nosetests

import os

from stetl.etl import ETL
from tests.stetl_test_case import StetlTestCase
from stetl.main import parse_args


class ConfigTest(StetlTestCase):
    """Basic configuration tests"""

    def setUp(self):
        super(ConfigTest, self).setUp()


        # Initialize Stetl
        self.curr_dir = os.path.dirname(os.path.realpath(__file__))
        self.cfg_dict = {'config_file': os.path.join(self.curr_dir, 'configs/copy_in_out_file.cfg')}

    def clear_stetl_env(self):
        # Restore old enviroment
        try:
            del os.environ['stetl_out_file']
            del os.environ['stetl_in_file']
        except:
            pass

    def tearDown(self):
        super(ConfigTest, self).tearDown()
        self.clear_stetl_env()

    def test_config_args_file_single(self):
        """
        Test single -a argsfile option
        :return:
        """
        args_default = os.path.join(self.curr_dir, 'configs/copy_in_out_file_default.args')
        args_parsed = parse_args(['-a', args_default])

        # Test args substitution from args_dict
        config_args = args_parsed.config_args
        self.assertEqual(config_args['in_file'], 'default_infile.txt')
        self.assertEqual(config_args['out_file'], 'default_outfile.txt')

    def test_config_args_explicit_single(self):
        """
        Test single -a "arg1=x arg2=y" option
        :return:
        """
        args_default = os.path.join(self.curr_dir, 'configs/copy_in_out_file_default.args')
        args_parsed = parse_args(['-a', 'in_file=default_infile.txt out_file=default_outfile.txt'])

        # Test args substitution from args_dict
        config_args = args_parsed.config_args
        self.assertEqual(config_args['in_file'], 'default_infile.txt')
        self.assertEqual(config_args['out_file'], 'default_outfile.txt')

    def test_config_args_file_multi(self):
        """
        Test multiple: -a argsfile1 -a argsfile2 option with override
        :return:
        """
        args_default = os.path.join(self.curr_dir, 'configs/copy_in_out_file_default.args')
        args_my = os.path.join(self.curr_dir, 'configs/copy_in_out_file_my.args')
        args_parsed = parse_args(['-a', args_default, '-a', args_my])

        # Test args substitution from args_dict
        config_args = args_parsed.config_args
        self.assertEqual(config_args['in_file'], 'my_infile.txt')
        self.assertEqual(config_args['out_file'], 'default_outfile.txt')

    def test_config_args_file_explicit_multi(self):
        """
        Test multiple: -a argsfile1 -a arg=myarg option with override
        :return:
        """
        args_default = os.path.join(self.curr_dir, 'configs/copy_in_out_file_default.args')
        args_parsed = parse_args(['-a', args_default, '-a', 'in_file=my_infile.txt'])

        # Test args substitution from args_dict
        config_args = args_parsed.config_args
        self.assertEqual(config_args['in_file'], 'my_infile.txt')
        self.assertEqual(config_args['out_file'], 'default_outfile.txt')

    def test_args_dict(self):
        args_dict = {'in_file': 'infile.txt', 'out_file': 'outfile.txt'}
        etl = ETL(self.cfg_dict, args_dict)

        # Test args substitution from args_dict
        self.assertEqual(etl.configdict.get('input_file', 'file_path'), 'infile.txt')
        self.assertEqual(etl.configdict.get('output_file', 'file_path'), 'outfile.txt')

    def test_args_dict_env_override(self):
        args_dict = {'in_file': 'infile.txt', 'out_file': 'outfile.txt'}

        # Override in OS env
        os.environ['stetl_in_file'] = 'env_infile.txt'

        etl = ETL(self.cfg_dict, args_dict)

        # Test args substitution from args_dict
        self.assertEqual(etl.configdict.get('input_file', 'file_path'), os.environ['stetl_in_file'])
        self.assertEqual(etl.configdict.get('output_file', 'file_path'), 'outfile.txt')

    def test_args_dict_env_all(self):
        """
        Substitute ALL args from OS env.
        :return:
        """

        # Set all args in in OS env
        os.environ['stetl_in_file'] = 'env_infile.txt'
        os.environ['stetl_out_file'] = 'env_outfile.txt'

        args_dict = None
        etl = ETL(self.cfg_dict, args_dict)

        # Test args substitution from args_dict
        self.assertEqual(etl.configdict.get('input_file', 'file_path'), os.environ['stetl_in_file'])
        self.assertEqual(etl.configdict.get('output_file', 'file_path'), os.environ['stetl_out_file'])



#!/usr/bin/env python
#
# Base classes for HTTP output like WFS-T and SOS-T or any other HTTP writing service.
#
# Author: Just van den Broecke
#
import httplib
import base64

from stetl.output import Output
from stetl.util import Util
from stetl.packet import FORMAT
from stetl.component import Config

log = Util.get_log('httpoutput')


class HttpOutput(Output):
    """
    Output via HTTP protocol, usually via POST.

    consumes=FORMAT.any
    """

    @Config(ptype=str, default=None, required=True)
    def host(self):
        """
        The hostname/IP addr for target request.
        """
        pass

    @Config(ptype=int, default=80, required=False)
    def port(self):
        """
        The port number for target request.
        """
        pass

    @Config(ptype=str, default='/', required=False)
    def path(self):
        """
        The path number for target request.
        """
        pass

    @Config(ptype=str, default='POST', required=False)
    def method(self):
        """
        The HTTP method for target request.
        """
        pass

    @Config(ptype=str, default='text/xml', required=False)
    def content_type(self):
        """
        The HTTP ContentType request header for target request.
        """
        pass

    @Config(ptype=str, default=None, required=False)
    def user(self):
        """
        The Username for HTTP basic auth for target request.
        """
        pass

    @Config(ptype=str, default=None, required=False)
    def password(self):
        """
        The Password for HTTP basic auth for target request.
        """
        pass

    @Config(ptype=bool, default=True, required=False)
    def list_fanout(self):
        """
        If we consume a list(), should we create a HTTP req for each member?
        """
        pass

    def __init__(self, configdict, section, consumes=FORMAT.any):
        Output.__init__(self, configdict, section, consumes)
        # self.accept_type = self.cfg.get('accept_type', self.content_type)
        self.req_nr = 0

    def create_payload(self, packet):
        """
        Create a HTTP body payload like for POST of an XML or JSON message.
        Subclasses like WFS and SOS override.
        :param packet:
        :return payload as string:
        """
        return packet.data

    def post(self, packet, payload):
        self.req_nr += 1

        webservice = httplib.HTTP(self.host, self.port)
        # write your headers
        webservice.putrequest(self.method, self.path)
        webservice.putheader("Host", self.host)
        webservice.putheader("User-Agent", "Stetl Python http")
        webservice.putheader("Content-Type", self.content_type)
        # webservice.putheader("Accept", self.accept_type)
        webservice.putheader("Content-Length", "%d" % len(payload))

        # basic auth: http://mozgovipc.blogspot.nl/2012/06/python-http-basic-authentication-with.html
        # base64 encode the username and password
        # write the Authorization header like: 'Basic base64encode(username + ':' + password)
        if self.user is not None:
            auth = base64.encodestring('%s:%s' % (self.user, self.password)).replace('\n', '')
            webservice.putheader("Authorization", "Basic %s" % auth)

        webservice.endheaders()
        webservice.send(payload)

        # get the response
        statuscode, statusmessage, header = webservice.getreply()
        log.info("Req nr %d - response status: code=%d msg=%s" % (self.req_nr, statuscode, statusmessage))
        if statuscode == 200:
            res = webservice.getfile().read()
        elif statuscode == 204:
            res = ''
        else:
            log.error("Headers: %s" % str(header))
            res = webservice.getfile().read()
            log.info('Content: %s' % res)

        return statuscode, statusmessage, res

    def write(self, packet):
        if packet.data is None:
            return packet

        if type(packet.data) is list and self.list_fanout is True:
            # Multiple records in list, save original
            original_data = packet.data
            for data_elm in original_data:
                packet.data = data_elm
                self.post(packet, self.create_payload(packet))

            packet.data = original_data

        else:
            # Regular, single data element or list_fanout is False
            self.post(packet, self.create_payload(packet))

        return packet



import mock
import os

from stetl.etl import ETL
from stetl.filters.fileextractor import ZipFileExtractor
from tests.stetl_test_case import StetlTestCase

class ZipFileExtractorTest(StetlTestCase):
    """Unit tests for ZipFileExtractor plus deprecation"""

    def setUp(self):
        super(ZipFileExtractorTest, self).setUp()

        # Initialize Stetl
        self.curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(self.curr_dir, 'configs/zipfileextractor.cfg')}
        self.etl = ETL(cfg_dict)
    
    def test_class(self):
        chain = StetlTestCase.get_chain(self.etl)
        section = StetlTestCase.get_section(chain, 1)
        class_name = self.etl.configdict.get(section, 'class')
        self.assertEqual('stetl.filters.fileextractor.ZipFileExtractor', class_name)
    
    def test_legacy_class(self):
        cfg_dict = {'config_file': os.path.join(self.curr_dir, 'configs/zipfileextractordeprecated.cfg')}
        self.etl = ETL(cfg_dict)
        chain = StetlTestCase.get_chain(self.etl)
        section = StetlTestCase.get_section(chain, 1)
        class_name = self.etl.configdict.get(section, 'class')

        # Deprecated class name in config
        self.assertEqual('stetl.filters.zipfileextractor.ZipFileExtractor', class_name)

        # Assigned to new version in fileextractor module!
        self.assertTrue(isinstance(chain.first_comp.next, ZipFileExtractor))

    def test_instance(self):
        chain = StetlTestCase.get_chain(self.etl)

        self.assertTrue(isinstance(chain.first_comp.next, ZipFileExtractor))
        
    @mock.patch('stetl.filters.fileextractor.ZipFileExtractor.after_chain_invoke', autospec=True)
    def test_execute(self, mock_after_chain_invoke):
        chain = StetlTestCase.get_chain(self.etl)
        chain.run()

        # ZIP file contains two GML files, both should be extracted; count is 3 because of final
        # call, so the ZipFileExtractor can indicate that no more files can be found.
        self.assertTrue(mock_after_chain_invoke.called)
        self.assertEqual(3, mock_after_chain_invoke.call_count)

        # Check if temp file exists
        section = StetlTestCase.get_section(chain, 1)
        file_path = self.etl.configdict.get(section, 'file_path')
        self.assertTrue(os.path.exists(file_path))
        os.remove(file_path)



# Lets data Packets pass-through, "sieve", based on criteria in their data.
# See issue: https://github.com/geopython/stetl/issues/78
#
# A concrete example is AttrValueRecordSieve which sieves records matching
# specific attribute values. One can also think of Sieves based on XPath expressions
# (e.g. for XML, GML), or geospatial, based on for example WFS-like filters like bounding boxes.
#
# Author: Just van den Broecke
#
from stetl.component import Config
from stetl.filter import Filter
from stetl.util import Util
from stetl.packet import FORMAT

log = Util.get_log('Sieve')


class Sieve(Filter):
    """
    ABC for specific Sieves that pass-through, "sieve",  Packets based on criteria in their data.
    """

    def __init__(self, configdict, section, consumes, produces):
        Filter.__init__(self, configdict, section, consumes, produces)

    def invoke(self, packet):
        if packet.data is None:
            return packet
        return self.sieve(packet)

    def sieve(self, packet):
        """
        To be implemented in subclasses.
        :param packet:
        :return:
        """
        return packet


class AttrValueRecordSieve(Sieve):
    """
    Sieves by attr/value(s) in Record Packets.
    """

    @Config(ptype=str, required=True)
    def attr_name(self):
        """
        Name of attribute whose value(s) are to be sieved.
        """
        pass

    @Config(ptype=list, default=list(), required=False)
    def attr_values(self):
        """
        Value(s) for attribute to be to sieved. If empty any value is passed through (existence
        of attr_name is criterium).
        """
        pass

    def __init__(self, configdict, section):
        Sieve.__init__(self, configdict, section, consumes=[FORMAT.record_array, FORMAT.record], produces=[FORMAT.record_array, FORMAT.record])

    def sieve(self, packet):
        """
        Filter out Packets that are not matching designated attr value(s).
        :param packet:
        :return:
        """

        # Start with empty result: fill with matching records
        record_data = packet.data
        packet.data = None

        # Data can be list or single record
        if type(record_data) is list:
            packet.data = list()
            for record in record_data:
                if self.matches_attr(record):
                    packet.data.append(record)
        elif type(record_data) is dict:
            if self.matches_attr(record_data):
                packet.data = record_data

        return packet

    def matches_attr(self, record):
        # Attr not even in record: no use going on
        if self.attr_name not in record:
            return False

        # Match if no value
        if not self.attr_values:
            return True

        return record[self.attr_name] in self.attr_values



# Output to File classes.
#
# Author: Just van den Broecke
#
import os

from stetl.output import Output
from stetl.util import Util
from stetl.packet import FORMAT
from stetl.component import Config

log = Util.get_log('fileoutput')


class FileOutput(Output):
    """
    Pretty print input to file. Input may be an etree doc or any other stringify-able input.

    consumes=FORMAT.any
    """

    # Start attribute config meta

    @Config(ptype=str, default=None, required=True)
    def file_path(self):
        """
        Path to file, for MultiFileOutput can be of the form like: gmlcities-%03d.gml
        """
        pass

    # End attribute config meta

    def __init__(self, configdict, section):
        Output.__init__(self, configdict, section, consumes=FORMAT.any)
        log.info("working dir %s" % os.getcwd())

    def write(self, packet):
        if packet.data is None:
            return packet

        return self.write_file(packet, self.file_path)

    def write_file(self, packet, file_path):
        log.info('writing to file %s' % file_path)
        out_file = open(file_path, 'w')

        out_file.write(packet.to_string())

        out_file.close()
        log.info("written to %s" % file_path)
        return packet


class MultiFileOutput(FileOutput):
    """
    Print to multiple files from subsequent packets like strings or etree docs, file_path must be of a form like: gmlcities-%03d.gml.

    consumes=FORMAT.any
    """

    def __init__(self, configdict, section):
        Output.__init__(self, configdict, section, consumes=FORMAT.any)
        self.file_num = 1

    def write(self, packet):
        if packet.data is None:
            return packet

        # file_path must be of the form: gmlcities-%03d.gml
        file_path = self.file_path % self.file_num
        self.file_num += 1
        return self.write_file(packet, file_path)



# -*- coding: utf-8 -*-
#
# Example of user-defined component.
#
# Author:Just van den Broecke

from stetl.util import Util
from stetl.filter import Filter
from stetl.packet import FORMAT

log = Util.get_log("myfilter")


class MyFilter(Filter):
    # Constructor
    def __init__(self, configdict, section):
        Filter.__init__(self, configdict, section, consumes=FORMAT.etree_doc, produces=FORMAT.etree_doc)

    def invoke(self, packet):
        if packet.data is None:
            return packet
        return self.do_something(packet)

    def do_something(self, packet):
        log.info("CALLING MyFilter OK!!!!")
        data = packet.data
        self.show_data(data)

        return packet

    def show_data(self, data):
        try:
            from lxml import etree
        except ImportError:
            # No lxml, so just print data
            print(data)
        else:
            root = data.getroot()
            for city in root.iterchildren():
                city_name = city.find('name').text
                print('Welcome to {c} !'.format(c=city_name))




import mock
import os

from stetl.etl import ETL
from stetl.filters.archiveexpander import ZipArchiveExpander
from tests.stetl_test_case import StetlTestCase

class ZipArchiveExpanderTest(StetlTestCase):
    """Unit tests for ZipArchiveExpander plus deprecation"""

    def setUp(self):
        super(ZipArchiveExpanderTest, self).setUp()

        # Initialize Stetl
        self.curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(self.curr_dir, 'configs/ziparchiveexpander.cfg')}
        self.etl = ETL(cfg_dict)

    def test_class(self):
        chain = StetlTestCase.get_chain(self.etl)
        section = StetlTestCase.get_section(chain, 2)
        class_name = self.etl.configdict.get(section, 'class')
        self.assertEqual('stetl.filters.archiveexpander.ZipArchiveExpander', class_name)

    def test_instance(self):
        chain = StetlTestCase.get_chain(self.etl)

        self.assertTrue(isinstance(chain.first_comp.next.next, ZipArchiveExpander))

    @mock.patch('stetl.outputs.standardoutput.StandardOutput.after_chain_invoke', autospec=True)
    def test_execute_zip(self, mock_after_chain_invoke):
        chain = StetlTestCase.get_chain(self.etl)
        chain.run()

        # ZIP file contains two GML files, both should be extracted; count is 3 because of final
        # call, so the ZipArchiveExpander can indicate that no more files can be found.
        self.assertTrue(mock_after_chain_invoke.called)
        self.assertEqual(2, mock_after_chain_invoke.call_count)

        # Check if temp dir exists
        section = StetlTestCase.get_section(chain, 2)
        target_dir = self.etl.configdict.get(section, 'target_dir')
        self.assertTrue(os.path.exists(target_dir))
        file_objects = os.listdir(target_dir)

        # 3 XML files in archive
        self.assertEqual(3, len(file_objects))

        for file_object in file_objects:
            file_object_path = os.path.join(target_dir, file_object)
            self.assertTrue(file_object.startswith('0221LIG'))
            self.assertTrue(file_object.endswith('.xml'))

            self.assertTrue(os.path.exists(file_object_path))
            os.remove(file_object_path)

    @mock.patch('stetl.filters.fileextractor.VsiFileExtractor.after_chain_invoke', autospec=True)
    def test_execute_vsizip(self, mock_after_chain_invoke):
        chain = StetlTestCase.get_chain(self.etl, index=1)
        chain.run()

        # ZIP file contains one XML file and is extracted; count is 2 because of final
        # call, so the VsiFileExtractor can indicate that no more files can be found.
        self.assertTrue(mock_after_chain_invoke.called)
        self.assertEqual(2, mock_after_chain_invoke.call_count)

        # Check if temp file exists
        section = StetlTestCase.get_section(chain, 2)
        target_dir = self.etl.configdict.get(section, 'target_dir')
        self.assertTrue(os.path.exists(target_dir))
        file_objects = os.listdir(target_dir)

        # 2 XML files in archive
        self.assertEqual(2, len(file_objects))

        for file_object in file_objects:
            file_object_path = os.path.join(target_dir, file_object)
            self.assertTrue(file_object.startswith('0221WPL'))
            self.assertTrue(file_object.endswith('.xml'))

            self.assertTrue(os.path.exists(file_object_path))
            os.remove(file_object_path)




#!/usr/bin/env python
#
# Executes the given command and returns the captured output.
#
# Author: Frank Steggink
#
import subprocess
import os
from stetl.component import Config
from stetl.filter import Filter
from stetl.util import Util
from stetl.packet import FORMAT

log = Util.get_log('execfilter')


class ExecFilter(Filter):
    """
    Executes any command (abstract base class).
    """

    @Config(ptype=str, default='', required=False)
    def env_args(self):
        """
        Provides of list of environment variables which will be used when executing the given command.

        Example: env_args = pgpassword=postgres othersetting=value~with~spaces
        """
        pass

    @Config(ptype=str, default='=', required=False)
    def env_separator(self):
        """
        Provides the separator to split the environment variable names from their values.
        """
        pass

    def __init__(self, configdict, section, consumes, produces):
        Filter.__init__(self, configdict, section, consumes, produces)

    def invoke(self, packet):
        return packet

    def execute_cmd(self, cmd):
        env_vars = Util.string_to_dict(self.env_args, self.env_separator)
        old_environ = os.environ.copy()

        try:
            os.environ.update(env_vars)
            log.info("executing cmd=%s" % cmd)
            result = subprocess.check_output(cmd, shell=True)
            log.info("execute done")
            return result
        finally:
            os.environ = old_environ


class CommandExecFilter(ExecFilter):
    """
    Executes an arbitrary command and captures the output

    consumes=FORMAT.string, produces=FORMAT.string
    """

    def __init__(self, configdict, section):
        ExecFilter.__init__(self, configdict, section, consumes=FORMAT.string, produces=FORMAT.string)

    def invoke(self, packet):
        if packet.data is not None:
            packet.data = self.execute_cmd(packet.data)

        return packet



# Output base class for ETL.
#
# Author: Just van den Broecke
#
from .component import Component
from .util import Util

log = Util.get_log('output')


class Output(Component):
    """
    Abstract Base class for all Output Components.

    """

    def __init__(self, configdict, section, consumes):
        Component.__init__(self, configdict, section, consumes=consumes, produces=None)
        log.info("cfg = %s" % self.cfg.to_string())

    def invoke(self, packet):
        packet = self.write(packet)
        packet.consume()
        return packet

    def write(self, packet):
        return packet



#!/usr/bin/env python
#
# Input classes for ETL.
#
# Author: Just van den Broecke
#
import codecs
import re

from stetl.component import Config
from stetl.postgis import PostGIS
from stetl.input import Input
from stetl.util import Util, etree, StringIO
from stetl.packet import FORMAT

log = Util.get_log('deegreeinput')


class DeegreeBlobstoreInput(Input):
    """
    Read features from deegree Blobstore DB into an etree doc.

    produces=FORMAT.etree_doc
    """

    # Start attribute config meta

    @Config(ptype=int, required=False, default=10000)
    def max_features_per_doc(self):
        """
        Max features to read from input feature GML stream per internal document.
        """
        pass

    @Config(ptype=str, required=True, default=None)
    def start_container(self):
        """
        Tag that starts container.
        """
        pass

    @Config(ptype=str, required=True, default=None)
    def end_container(self):
        """
        Tag that ends container.
        """
        pass

    @Config(ptype=str, required=False, default=False)
    def start_feature_tag(self):
        """
        XML tag that starts Feature.
        """
        pass

    @Config(ptype=str, required=False, default=None)
    def end_feature_tag(self):
        """
        XML tag that ends Feature.
        """
        pass

    # End attribute config meta

    def __init__(self, configdict, section):
        Input.__init__(self, configdict, section, produces=FORMAT.etree_doc)
        self.cur_feature_blob = None
        self.rowcount = 0

        # http://www.mkyong.com/regular-expressions/how-to-extract-html-links-with-regular-expression/
        self.regex_xlink_href = re.compile("\\s*(?i)xlink:href\\s*=\\s*(\"#([^\"]*\")|'#[^']*'|(#[^'\">\\s]+))")

        self.db = None
        self.xlink_db = None
        self.buffer = None
        self.feature_count = 0
        # Reusable XML parser
        self.xml_parser = etree.XMLParser(remove_blank_text=True)

    def init(self):
        pass

    def read(self, packet):
        if packet.is_end_of_stream():
            return packet

        if self.db is None:
            # First time read
            log.info("reading records from blobstore..")
            self.db = PostGIS(self.cfg.get_dict())
            self.db.connect()
            sql = self.cfg.get('sql')
            self.rowcount = self.db.execute(sql)
            self.cur = self.db.cursor
            log.info("Read records rowcount=%d" % self.rowcount)

            # Init separate connection to fetch objects referenced by xlink:href
            self.xlink_db = PostGIS(self.cfg.get_dict())
            self.xlink_db.connect()

        # Query active
        while self.cur is not None:
            if self.buffer is None:
                self.buffer = self.init_buf()
                self.buffer.write(self.start_container)

            # Get next blob record
            record = self.cur.fetchone()

            # End of all records
            if record is None:
                # End of records: start closing
                self.buffer.write(self.end_container)
                self.cur = None
                self.db.commit()

                # Only create doc if there are features in the buffer
                if self.feature_count > 0:
                    self.buffer_to_doc(packet)
                packet.set_end_of_doc()
                break
            else:
                # New record: embed feature blob in feature tags and write to buffer
                feature_blob = self.write_feature(record)

                # If we have local xlinks: fetch the related features as well from the DB and
                # output them within the same document (local href resolvable)
                # TODO: in some cases we may need to be recursive (xlinks in xlinked features...)

                # First construct a single query for all xlinks
                xlink_sql = None
                for xlink in self.regex_xlink_href.finditer(feature_blob):
                    gml_id = xlink.group(1).strip('"').strip('#')
                    # We don't want multiple occurences of the same xlinked feature
                    if gml_id in self.xlink_ids:
                        continue

                    self.xlink_ids.add(gml_id)
                    if xlink_sql is None:
                        xlink_sql = "SELECT binary_object from gml_objects where gml_id = '%s'" % gml_id
                    else:
                        xlink_sql += "OR gml_id = '%s'" % gml_id

                # Should we retrieve and write xlinked features?
                if xlink_sql is not None:
                    # Fetch from DB
                    self.xlink_db.execute(xlink_sql)
                    while True:
                        # Get next blob record
                        xlink_record = self.xlink_db.cursor.fetchone()
                        if xlink_record is None:
                            break
                        self.write_feature(xlink_record)

                # Should we output a doc
                if self.feature_count >= self.max_features_per_doc:
                    # End of records: create XML doc
                    self.buffer.write(self.end_container)
                    self.buffer_to_doc(packet)
                    break

        if self.cur is None:
            # All records handled: close off
            packet.set_end_of_stream()
            # log.info("[%s]" % packet.data)

        return packet

    def write_feature(self, record):
        feature_blob = str(record[0])

        # Write start-tag, blob element, end-tag
        self.buffer.write(self.start_feature_tag)
        self.buffer.write(feature_blob)
        self.buffer.write(self.end_feature_tag)
        self.feature_count += 1
        return feature_blob

    def init_buf(self):
        buffer = StringIO()
        buffer = codecs.getwriter("utf8")(buffer)
        self.feature_count = 0
        self.xlink_ids = set()
        return buffer

    def buffer_to_doc(self, packet):
        # Process/transform data in buffer
        self.buffer.seek(0)
        try:
            packet.data = etree.parse(self.buffer, self.xml_parser)
        except Exception as e:
            bufStr = self.buffer.getvalue()
            if not bufStr:
                log.info("parse buffer empty: content=[%s]" % bufStr)
            else:
                log.error("error in buffer parsing %s" % str(e))
                raise
        self.buffer.close()
        self.buffer = None



# Output classes for ETL, databases.
#
# Author: Just van den Broecke
#
from stetl.output import Output
from stetl.util import Util
from stetl.packet import FORMAT
from stetl.component import Config
from stetl.postgis import PostGIS

log = Util.get_log('dboutput')


class DbOutput(Output):
    """
    Output to any database (abstract base class).
    """

    def __init__(self, configdict, section, consumes):
        Output.__init__(self, configdict, section, consumes)

    def write(self, packet):
        return packet


class PostgresDbOutput(DbOutput):
    """
    Output to PostgreSQL database.
    Input is an SQL string.
    Output by executing input SQL string.

    consumes=FORMAT.string
    """

    # Start attribute config meta
    @Config(ptype=str, required=True, default=None)
    def database(self):
        """
        Database name.
        """
        pass

    @Config(ptype=str, required=False, default=None)
    def user(self):
        """
        DB User name.
        """
        pass

    @Config(ptype=str, required=False, default=None)
    def password(self):
        """
        DB Password for user.
        """
        pass

    @Config(ptype=str, required=False, default=None)
    def host(self):
        """
        Hostname for DB.
        """
        pass

    @Config(ptype=str, required=False, default='public')
    def schema(self):
        """
        Postgres schema name for DB.
        """
        pass

    # End attribute config meta

    def __init__(self, configdict, section):
        DbOutput.__init__(self, configdict, section, consumes=FORMAT.string)

    def write(self, packet):
        if packet.data is None:
            return packet

        log.info('executing SQL')
        db = PostGIS(self.cfg.get_dict())
        rowcount = db.tx_execute(packet.data)
        log.info('executed SQL, rowcount=%d' % rowcount)
        return packet


class PostgresInsertOutput(PostgresDbOutput):
    """
    Output by inserting a single record in a Postgres database table.
    Input is a Stetl record (Python dict structure) or a list of records.
    Creates an INSERT for Postgres to insert each single record.
    When the "replace" parameter is True, any existing record keyed by "key" is
    attempted to be UPDATEd first.

    NB a constraint is that the first and each subsequent each record needs to contain
    all values as an INSERT and UPDATE query template is built once for the columns
    in the first record.

    consumes=[FORMAT.record_array, FORMAT.record]
    """

    # Start attribute config meta
    @Config(ptype=str, required=False, default='public')
    def table(self):
        """
        Table for inserts.
        """
        pass

    @Config(ptype=bool, required=False, default=False)
    def replace(self):
        """
        Replace record if exists?
        """
        pass

    @Config(ptype=str, required=False, default=None)
    def key(self):
        """
        The key column name of the table, required when replacing records.
        """
        pass

    # End attribute config meta

    def __init__(self, configdict, section, consumes=FORMAT.record):
        DbOutput.__init__(self, configdict, section, consumes=[FORMAT.record_array, FORMAT.record])
        self.query = None
        self.update_query = None
        self.db = None

    def init(self):
        # Connect only once to DB
        log.info('Init: connect to DB')
        self.db = PostGIS(self.cfg.get_dict())
        self.db.connect()

    def exit(self):
        # Disconnect from DB when done
        log.info('Exit: disconnect from DB')
        self.db.disconnect()

    def create_query(self, record):
        # We assume that all records do the same INSERT key/values
        # See http://grokbase.com/t/postgresql/psycopg/12735bvkmv/insert-into-with-a-dictionary-or-generally-with-a-variable-number-of-columns
        # e.g. INSERT INTO lml_files ("file_name", "file_data") VALUES (%s,%s)
        query = "INSERT INTO %s (%s) VALUES (%s)" % (
            self.cfg.get('table'), ",".join(['%s' % k for k in record]), ",".join(["%s", ] * len(record.keys())))
        log.info('query is %s', query)
        return query

    def create_update_query(self, record):
        # We assume that all records do the same UPDATE key/values
        # https://stackoverflow.com/questions/1109061/insert-on-duplicate-update-in-postgresql/6527838#6527838
        # e.g. UPDATE table SET field='C', field2='Z' WHERE id=3;
        query = "UPDATE %s SET (%s) = (%s) WHERE  %s = %s" % (
            self.cfg.get('table'), ",".join(['%s ' % k for k in record]), ",".join(["%s", ] * len(record.keys())), self.key, "%s")
        log.info('update query is %s', query)
        return query

    def insert(self, record):
        res = 0
        if self.replace and self.key and self.key in record:

            # Replace option: try UPDATE if existing
            # https://stackoverflow.com/questions/1109061/insert-on-duplicate-update-in-postgresql/6527838#6527838
            values = list(record.values())
            values.append(record[self.key])
            res = self.db.execute(self.update_query, values)
            # del_query = "DELETE FROM %s WHERE %s = '%s'" % (self.cfg.get('table'), self.key, record[self.key])
            # res = self.db.execute(del_query)

        if res < 1:
            # Do insert with values from the record dict
            # only if we did not do an UPDATE (res==0) on existing record.
            self.db.execute(self.query, list(record.values()))
        self.db.commit(close=False)

    def write(self, packet):
        # Deal with empty or zero-length data structures (list or dict)
        if packet.data is None or len(packet.data) == 0:
            return packet

        # ASSERT: record data present

        # record is Python dict (single record) or list of Python dict (multiple records)
        record = packet.data

        # Generate INSERT query template once
        first_record = record
        if type(record) is list and len(record) > 0:
            first_record = record[0]

        # Create INSERT and optional UPDATE query-templates once
        if self.query is None:
            self.query = self.create_query(first_record)

        if self.replace and self.key and not self.update_query:
            self.update_query = self.create_update_query(first_record)

        # Check if record is single (dict) or array (list of dict)
        if type(record) is dict:
            # Do insert with values from the single record
            self.insert(record)

            # log.info('committed record key=%s' % record[self.key])

        elif type(record) is list:
            # Multiple records in list
            for rec in record:
                # Do insert with values from the record
                self.insert(rec)

            log.info('committed %d records' % len(record))

        return packet



# testing: to be called by nosetests

import os

from stetl.etl import ETL
from tests.stetl_test_case import StetlTestCase


class ConfigTest(StetlTestCase):
    """Basic configuration tests"""

    def setUp(self):
        super(ConfigTest, self).setUp()

        # Initialize Stetl
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(curr_dir, 'configs/copy_in_out.cfg')}
        self.etl = ETL(cfg_dict)

    def test_type(self):
        self.assertEqual(self.etl.configdict.get('etl', 'chains'), 'input_xml_file|output_std')

    def test_run(self):
        self.etl.run()



#!/usr/bin/env python
#
# Output Components for deegree server storage (www.deegree.org).
#
# Author: Just van den Broecke
#
# NB deegree also supports WFS-T!
#
import os
from stetl.postgis import PostGIS
from stetl.output import Output
from stetl.util import Util, etree
from stetl.packet import FORMAT

log = Util.get_log('deegreeoutput')


class DeegreeBlobstoreOutput(Output):
    """
    Insert features into deegree Blobstore from an etree doc.

    consumes=FORMAT.etree_doc
    """

    def __init__(self, configdict, section):
        Output.__init__(self, configdict, section, consumes=FORMAT.etree_doc)
        self.overwrite = self.cfg.get_bool('overwrite')
        self.srid = self.cfg.get_int('srid', -1)
        self.feature_member_tag = self.cfg.get('feature_member_tag')
        self.feature_type_ids = {}

    def init(self):
        self.get_feature_types()
        if self.overwrite:
            self.delete_features()
            # not required for now
            self.pg_srs_constraint()

    def get_feature_types(self):
        log.info('reading all featuretypes from DB')
        db = PostGIS(self.cfg.get_dict())
        db.connect()
        sql = "SELECT id,qname FROM feature_types"
        db.execute(sql)
        cur = db.cursor
        for record in cur:
            self.feature_type_ids[record[1]] = record[0]

    def delete_features(self):
        log.info('deleting ALL features in DB')
        db = PostGIS(self.cfg.get_dict())
        db.tx_execute("TRUNCATE gml_objects")

    def pg_srs_constraint(self):
        log.info('set srs constraint')
        db = PostGIS(self.cfg.get_dict())
        srid = self.srid
        sql = "ALTER TABLE gml_objects DROP CONSTRAINT enforce_srid_gml_bounded_by;"
        db.tx_execute(sql)
        sql = "ALTER TABLE gml_objects ADD CONSTRAINT enforce_srid_gml_bounded_by CHECK  (st_srid(gml_bounded_by) = (%s));" % srid
        db.tx_execute(sql)

    def write(self, packet):
        if packet.data is None:
            return packet

        gml_doc = packet.data
        log.info('inserting features in DB')
        db = PostGIS(self.cfg.get_dict())
        db.connect()
        #        print (self.to_string(gml_doc, False, False))
        #        NS = {'base': 'urn:x-inspire:specification:gmlas:BaseTypes:3.2', 'gml': 'http://www.opengis.net/gml/3.2'}
        #        featureMembers = gml_doc.xpath('//base:member/*', namespaces=NS)
        featureMembers = gml_doc.xpath("//*[local-name() = '%s']/*" % self.feature_member_tag)
        count = 0
        gml_ns = None
        for childNode in featureMembers:
            if gml_ns is None:
                if 'gml' in childNode.nsmap:
                    gml_ns = childNode.nsmap['gml']
                elif 'GML' in childNode.nsmap:
                    gml_ns = childNode.nsmap['GML']

            gml_id = childNode.get('{%s}id' % gml_ns)

            feature_type_id = self.feature_type_ids[childNode.tag]

            # Find a GML geometry in the GML NS
            # ogrGeomWKT = None
            #            gmlMembers = childNode.xpath(".//gml:Point|.//gml:Curve|.//gml:Surface|.//gml:MultiSurface", namespaces=NS)
            gmlMembers = childNode.xpath(
                ".//*[local-name() = 'Point']|.//*[local-name() = 'Polygon']|.//*[local-name() =\
                 'Curve']|.//*[local-name() = 'Surface']|.//*[local-name() = 'MultiSurface']")
            geom_str = None
            for gmlMember in gmlMembers:
                if geom_str is None:
                    geom_str = etree.tostring(gmlMember)
                    #                   no need for GDAL Python bindings for now, maybe when we'll optimize with COPY iso INSERT
            # ogrGeom = ogr.CreateGeometryFromGML(str(gmlStr))
            #                    if ogrGeom is not None:
            #                        ogrGeomWKT = ogrGeom.ExportToWkt()
            #                        if ogrGeomWKT is not None:
            #                            break

            blob = etree.tostring(childNode, pretty_print=False, xml_declaration=False, encoding='UTF-8')

            if geom_str is None:
                sql = "INSERT INTO gml_objects(gml_id, ft_type, binary_object) VALUES (%s, %s, %s)"
                parameters = (gml_id, feature_type_id, db.make_bytea(blob))
            else:
                # ST_SetSRID(ST_GeomFromGML(%s)),-1)
                sql = "INSERT INTO gml_objects(gml_id, ft_type, binary_object, gml_bounded_by) VALUES (%s, %s, %s, ST_SetSRID( ST_GeomFromGML(%s),%s) )"
                parameters = (gml_id, feature_type_id, db.make_bytea(blob), geom_str, self.srid)

            if db.execute(sql, parameters) == -1:
                log.error("feat num# = %d error inserting feature blob=%s (but continuing)" % (count, blob))

                # will fail but we will close connection also
                db.commit()

                # proceed...
                log.info('retrying to proceed with remaining features...')
                db = PostGIS(self.cfg.get_dict())
                db.connect()
                count = 0

            count += 1

        exception = db.commit()
        if exception is not None:
            log.error("error in commit")

        log.info("inserted %s features" % count)
        return packet


#
class DeegreeFSLoaderOutput(Output):
    """
    Insert features via deegree using deegree's FSLoader tool from an etree doc.

    consumes=FORMAT.etree_doc
    """
    # d3toolbox FeatureStoreLoader -action insert -dataset ${DATASET} -format ${GML_VERSION} -fsconfig ${1} -idgen ${IDGEN_MODE} -workspace ${WORKSPACE}

    cmd_tmpl = '%s|FeatureStoreLoader|-action|insert|-dataset|%s|-format|%s|-fsconfig|%s|-idgen|%s|-workspace|%s'

    def __init__(self, configdict, section):
        Output.__init__(self, configdict, section, consumes=FORMAT.etree_doc)

    def write(self, packet):
        from subprocess import Popen, PIPE

        if packet.data is None:
            return packet

        d3tools_path = self.cfg.get('d3tools_path')
        workspace_path = self.cfg.get('workspace_path')
        feature_store = self.cfg.get('feature_store')
        format = self.cfg.get('format', 'GML_32')
        idgen = self.cfg.get('idgen', 'USE_EXISTING')
        java_opts = self.cfg.get('java_opts', "-Xms128m -Xmx1024m")
        dataset = self.cfg.get('dataset_path', '/dev/stdin')
        os.putenv("JAVA_OPTS", java_opts)

        d3tools = d3tools_path + '/bin/d3toolbox'
        cmd = DeegreeFSLoaderOutput.cmd_tmpl % (d3tools, dataset, format, feature_store, idgen, workspace_path)
        cmd = cmd.split('|')

        p = Popen(cmd, stdin=PIPE)

        p.stdin.write(packet.to_string())

        # result = p.communicate()[0]
        # print(result)
        return packet



# PostGIS support wrapper.
#
# Author: Just van den Broecke
#
from .util import Util

log = Util.get_log("postgis")

try:
    import psycopg2
    import psycopg2.extensions
except ImportError:
    log.error("cannot find package psycopg2 for Postgres client support, please install psycopg2 first!")
    # sys.exit(-1)


class PostGIS:
    def __init__(self, config):
        # Lees de configuratie
        self.config = config
        self.e = None

    def initialiseer(self, bestand):
        log.info('Connecting...')
        self.connect(True)

        log.info('executing sql script...')
        try:
            script = open(bestand, 'r').read()
            self.cursor.execute(script)
            self.connection.commit()
            log.info('script executed')
        except psycopg2.DatabaseError as e:
            log.warn("error '%s' from script '%s'" % (str(e), str(bestand)))

    def connect(self, initdb=False):
        try:
            conn_str = "dbname=%s user=%s host=%s port=%s" % (self.config['database'],
                                                              self.config['user'],
                                                              self.config.get('host', 'localhost'),
                                                              self.config.get('port', '5432'))
            log.info('Connecting to %s' % conn_str)
            conn_str += ' password=%s' % self.config['password']
            self.connection = psycopg2.connect(conn_str)
            self.cursor = self.connection.cursor()

            self.set_schema()
            log.debug("Connected to database %s" % (self.config['database']))
        except Exception as e:
            log.error("Cannot connect to database '%s' e=%s" % (self.config['database'], str(e)))

    def disconnect(self):
        self.e = None
        try:
            self.connection.close()
        except Exception as e:
            self.e = e
            log.error("error %s in close" % (str(e)))

        return self.e

    def commit(self, close=True):
        self.e = None
        try:
            self.connection.commit()
            if close is True:
                self.connection.close()
        except Exception as e:
            self.e = e
            log.error("error %s in commit" % (str(e)))

        return self.e

    def create_schema(self):
        # Public schema: no further action required
        if self.config['schema'] != 'public':
            # A specific schema is required create it and set the search path
            self.execute('''DROP SCHEMA IF EXISTS %s CASCADE;''' % self.config['schema'])
            self.execute('''CREATE SCHEMA %s;''' % self.config['schema'])
            self.connection.commit()

    def set_schema(self):
        # Non-public schema set search path
        if self.config['schema'] != 'public':
            # Always set search path to our schema
            self.execute('SET search_path TO %s,public' % self.config['schema'])
            self.connection.commit()

    def log_action(self, action, bestand="n.v.t", bericht='geen', error=False):
        sql = "INSERT INTO stetl_log(actie, bestand, error, bericht) VALUES (%s, %s, %s, %s)"
        parameters = (action, bestand, error, bericht)
        self.tx_execute(sql, parameters)

    def log_meta(self, key, value):
        sql = "INSERT INTO stetl_info(sleutel, waarde) VALUES (%s, %s)"
        parameters = (key, value)
        self.tx_execute(sql, parameters)

    def make_bytea(self, blob):
        return psycopg2.Binary(blob)

    def get_column_names(self, table, schema='public'):
        self.cursor.execute(
            "select column_name from information_schema.columns where table_schema = '%s' and table_name='%s'" % (schema, table))
        column_names = [row[0] for row in self.cursor]
        return column_names

    def execute(self, sql, parameters=None):
        try:
            if parameters:
                self.cursor.execute(sql, parameters)
            else:
                self.cursor.execute(sql)

                # log.debug(self.cursor.statusmessage)
        except Exception as e:
            log.error("error %s in query: %s with params: %s" % (str(e), str(sql), str(parameters)))
            #            self.log_actie("uitvoeren_db", "n.v.t", "fout=%s" % str(e), True)
            return -1

        return self.cursor.rowcount

    def file_execute(self, sqlfile):
        self.e = None
        try:
            log.info("Executing SQL of file = %s ..." % sqlfile)
            self.connect()
            f = open(sqlfile, 'r')
            sql = f.read()
            self.execute(sql)
            self.connection.commit()
            f.close()
            log.info("SQL executed OK")
        except Exception as e:
            self.e = e
            self.log_action("file_execute", "n.v.t", "fout=%s" % str(e), True)
            log.warn("can't execute SQL script, error: %s" % (str(e)))

    def tx_execute(self, sql, parameters=None):
        self.e = None
        try:
            self.connect()
            self.execute(sql, parameters)
            self.connection.commit()
            self.connection.close()

            # log.debug(self.cursor.statusmessage)
        except Exception as e:
            self.e = e
            log.error("error %s in transaction: %s with parms: %s" % (str(e), str(sql), str(parameters)))
            # Return a dummy "-1" row count in the same vein as self.execute()
            return -1

        # Make sure there is a cursor with a row count before returning it, otherwise "return early"
        if not hasattr(self, 'cursor'):
            return -1

        if not hasattr(self.cursor, 'rowcount'):
            return -1

        return self.cursor.rowcount



# Chain class: holds pipeline of components.
#
# Author: Just van den Broecke
#

from .factory import factory
from .packet import Packet
from .util import Util
from .splitter import Splitter
from .merger import Merger

log = Util.get_log('chain')


class Chain:
    """
    Holder for single invokable pipeline of components
    A Chain is basically a singly linked list of Components
    Each Component executes a part of the total ETL.
    Data along the Chain is passed within a Packet object.
    The compatibility of input and output for linked
    Components is checked when adding a Component to the Chain.
    """

    def __init__(self, chain_str, config_dict):
        self.first_comp = None
        self.cur_comp = None
        self.config_dict = config_dict
        self.chain_str = chain_str.strip()

    def assemble(self):
        """
        Builder method: build a Chain of linked Components
        :return:
        """
        log.info('Assembling Chain: %s...' % self.chain_str)

        # Create linked list of input/filter/output (ETL Component) objects
        chain_str = self.chain_str
        sub_comps = []
        while chain_str:
            chain_str = chain_str.strip()

            # Check and handle Splitter construct
            # e.g. input_xml_file |(transformer_xslt|output_file) (output_std) (transformer_xslt|output_std)
            if chain_str.startswith('('):
                etl_section_name, chain_str = chain_str.split(')', 1)
                etl_section_name = etl_section_name.strip('(')

                # Check for subchain (split at Filter level)
                if '|' in etl_section_name:
                    # Have subchain: use Chain to assemble
                    sub_chain = Chain(etl_section_name, self.config_dict)
                    sub_chain.assemble()
                    child_comp = sub_chain.first_comp
                else:
                    # Single component (Output) to split
                    child_comp = factory.create_obj(self.config_dict, etl_section_name.strip())

                # Assemble Components (can be subchains) for Splitter later
                sub_comps.append(child_comp)
                if '(' in chain_str:
                    # Still components (subchains) to assemble for Splitter
                    continue

            if len(sub_comps) > 0:
                if chain_str.startswith('|'):
                    # Next component is Merger with children
                    etl_comp = Merger(self.config_dict, sub_comps)
                    dummy, chain_str = chain_str.split('|', 1)
                else:
                    # Next component is Splitter with children
                    etl_comp = Splitter(self.config_dict, sub_comps)
                sub_comps = []
            else:

                # "Normal" case: regular Components piped in Chain
                if '|' in chain_str:
                    # More than one component in remaining Chain
                    etl_section_name, chain_str = chain_str.split('|', 1)
                else:
                    # Last element, we're done!
                    etl_section_name = chain_str
                    chain_str = None

                # Create the ETL component by name and properties
                etl_comp = factory.create_obj(self.config_dict, etl_section_name.strip())

            # Add component to end of Chain
            self.add(etl_comp)

    def add(self, etl_comp):
        """
        Add component to end of Chain
        :param etl_comp:
        :return:
        """
        if not self.first_comp:
            self.first_comp = etl_comp
        else:
            # Already component(s) in chain add to current
            self.cur_comp.add_next(etl_comp)

        # Remember current
        self.cur_comp = etl_comp

    def get_by_class(self, clazz):
        """
        Get Component instance from Chain by class, mainly for testing.
        :param clazz:
        :return Component:
        """
        cur_comp = self.first_comp
        while cur_comp:
            if cur_comp.__class__ == clazz:
                return cur_comp

            # Try next in Chain
            cur_comp = cur_comp.next

        return None

    def get_by_id(self, id):
        """
        Get Component instance from Chain, mainly for testing.
        :param name:
        :return Component:
        """
        cur_comp = self.first_comp
        while cur_comp:
            if cur_comp.get_id() == id:
                return cur_comp

            # Try next in Chain
            cur_comp = cur_comp.next

        return None

    def get_by_index(self, index):
        """
        Get Component instance from Chain by position/index in Chain, mainly for testing.
        :param clazz:
        :return Component:
        """
        cur_comp = self.first_comp
        i = 0
        while cur_comp and i < index:
            # Try next in Chain
            cur_comp = cur_comp.next
            i += 1

        return cur_comp

    def run(self):
        """
        Run the ETL Chain.
        :return:
        """
        log.info('Running Chain: %s' % self.chain_str)

        # One time init for entire Chain
        self.first_comp.do_init()

        # Do ETL as long as input available in Packet
        packet = Packet()
        rounds = 0
        try:
            while not packet.is_end_of_stream():
                # try:
                # Invoke the first component to start the chain
                packet.init()
                packet = self.first_comp.process(packet)
                rounds += 1
                #            except (Exception), e:
                #                log.error("Fatal Error in ETL: %s"% str(e))
                #                break
        finally:
            # Always one time exit for entire Chain
            self.first_comp.do_exit()

        log.info('DONE - %d rounds - chain=%s ' % (rounds, self.chain_str))



# Merger Component base class for ETL.
#
# Author: Just van den Broecke
#

import random
from .util import Util
from .component import Component

log = Util.get_log('merger')


class Merger(Component):
    """
    Component that merges multiple Input Components into a single Component.
    Use this for example to combine multiple input streams like API endpoints.
    The Merger will embed Child Components to which actions are delegated.
    A Child Component may be a sub-Chain e.g. (Input|Filter|Filter..) sequence.
    Hence the "next" should be coupled to the last Component in that sub-Chain with
    the degenerate case where the sub-Chain is a single (Input) Component.
    NB this Component can only be used for Inputs.
    """

    def __init__(self, config_dict, child_list):
        # Assemble child list
        self.children = []
        section_name = ''
        for child in child_list:
            section_name += '-%s_%d' % (child.get_id(), random.randrange(0, 100000))

            # A Child can be a sub-Chain: each child is tuple: [0] is first
            # [1] is last in sub-Chain. child[0] === child[1] if child is single Component.
            # Need to remember both first and last in order to link/unlink subchain.
            # So we store the Child as a tuple of (first, last).
            self.children.append((child, child.get_last()))

        # Add ourselves to config for compat with Component
        config_dict.add_section(section_name)

        # We use the in/out formats of first child, will be compat-checked later
        Component.__init__(self, config_dict, section_name, consumes=self.first(self.children[0])._input_format,
                           produces=self.last(self.children[0])._output_format)

        self.end_count = len(self.children)

    def add_next(self, next_component):
        for child in self.children:
            # Couple Child Component's last .next directly to our next
            self.last(child).add_next(next_component)

        # Remember next
        self.next = next_component

    def first(self, child):
        """
        Get first Component in Child sub-Chain.
        :param child:
        :return: first Component
        """
        return child[0]

    def last(self, child):
        """
        Get last Component in Child sub-Chain.
        :param child:
        :return: last Component
        """
        return child[1]

    # Check compatibility with our child Components
    def is_compatible(self):
        for child in self.children:
            # Last in subchain must be compatible
            if not self.last(child).is_compatible():
                return False
        return True

    def process(self, packet):
        # Defer processing to children
        # and track of End-of-Stream Packets

        for child in self.children:
            # Skip inactive child Components
            if not self.last(child).next:
                continue

            # Defer to child
            self.first(child).process(packet)

            # Keep track of End-of-Stream
            if packet.is_end_of_stream():
                # deactivate Child by unlinking
                # otherwise we'll keep getting EoS
                self.last(child).next = None
                self.end_count -= 1

            # Re-init to start afresh again
            packet.init()

        # Only if all children have End-of-Stream
        # declare the Packet returned EoS.
        if self.end_count == 0:
            packet.set_end_of_stream()

        return packet

    def do_init(self):
        for child in self.children:
            # Only init the child, without
            # initing upstream Components via Chain
            self.last(child).next = None
            self.first(child).do_init()
            self.last(child).next = self.next

        # init upstream Components once
        self.next.do_init()

    def do_exit(self):
        for child in self.children:
            # Only exit the child, without
            # exiting upstream Components via Chain
            self.last(child).next = None
            self.first(child).do_exit()

        # exit upstream Components once
        self.next.do_exit()



# Output classes for ETL.
#
# Author: Just van den Broecke
#
from stetl.output import Output
from stetl.util import Util
from stetl.packet import FORMAT

log = Util.get_log('standardoutput')


#
class StandardOutput(Output):
    """
    Print any input to standard output.

    consumes=FORMAT.any
    """

    def __init__(self, configdict, section):
        Output.__init__(self, configdict, section, consumes=FORMAT.any)

    def write(self, packet):
        if packet.data is None:
            return packet

        # Default: print to stdout
        print(packet.to_string())
        return packet


class StandardXmlOutput(Output):
    """
    Pretty print XML from etree doc to standard output. OBSOLETE, can be done with  StandardOutput

    consumes=FORMAT.etree_doc
    """

    def __init__(self, configdict, section):
        Output.__init__(self, configdict, section, consumes=FORMAT.etree_doc)

    def write(self, packet):
        if packet.data is None:
            return packet

        # Default: print to stdout
        print(packet.to_string())
        return packet



import mock
import os
import sys

from stetl.etl import ETL
from stetl.outputs.execoutput import CommandExecOutput
from tests.stetl_test_case import StetlTestCase

class CommandExecOutputTest(StetlTestCase):
    """Unit tests for CommandExecOutput"""

    def setUp(self):
        super(CommandExecOutputTest, self).setUp()

        # Initialize Stetl
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(curr_dir, 'configs/commandexecoutput.cfg')}
        self.etl = ETL(cfg_dict)
    
    def test_class(self):
        chain = StetlTestCase.get_chain(self.etl)
        section = StetlTestCase.get_section(chain, -1)
        class_name = self.etl.configdict.get(section, 'class')
        
        self.assertEqual('stetl.outputs.execoutput.CommandExecOutput', class_name)
    
    def test_instance(self):
        chain = StetlTestCase.get_chain(self.etl)

        self.assertTrue(isinstance(chain.cur_comp, CommandExecOutput))
    
    @mock.patch('subprocess.call', autospec=True)
    def test_execute(self, mock_call):
        # Read content of input file
        chain = StetlTestCase.get_chain(self.etl)
        section = StetlTestCase.get_section(chain)
        fn = self.etl.configdict.get(section, 'file_path')
        with open(fn, 'r') as f:
            contents = f.read()

        self.etl.run()
        
        self.assertTrue(mock_call.called)
        self.assertEqual(1, mock_call.call_count)
        args, kwargs = mock_call.call_args
        self.assertEqual(contents, args[0])
        



# Packet buffering.
#
# Author:Just van den Broecke

import copy
from stetl.util import Util
from stetl.filter import Filter
from stetl.packet import FORMAT

log = Util.get_log("packetbuffer")


class PacketBuffer(Filter):
    """
    Buffers all incoming Packets, main use is unit-testing to inspect Packets after ETL is done.
    """

    # Constructor
    def __init__(self, configdict, section):
        Filter.__init__(self, configdict, section, consumes=FORMAT.any, produces=FORMAT.any)
        self.packet_list = []

    def invoke(self, packet):
        # Buffer Packet and pass-through, we need a deep copy as Packets may be cleared/reused
        self.packet_list.append(copy.copy(packet))
        return packet



#!/usr/bin/env python
#
# Extracts data from a string using a regular expression and generates a record.
#
# Author: Frank Steggink
import re

from stetl.component import Config
from stetl.filter import Filter
from stetl.packet import FORMAT
from stetl.util import Util

log = Util.get_log("regexfilter")


class RegexFilter(Filter):
    """
    Extracts data from a string using a regular expression and returns the named groups as a record.
    consumes=FORMAT.string, produces=FORMAT.record
    """

    # Start attribute config meta
    # Applying Decorator pattern with the Config class to provide
    # read-only config values from the configured properties.

    @Config(ptype=str, default=None, required=True)
    def pattern_string(self):
        """
        Regex pattern string. Should contain named groups.
        """
        pass

    # End attribute config meta

    # Constructor
    def __init__(self, configdict, section, consumes=FORMAT.string, produces=FORMAT.record):
        Filter.__init__(self, configdict, section, consumes, produces)

        self.regex_object = re.compile(self.pattern_string, re.S)

    def init(self):
        log.info('Init: regex filter')
        if self.pattern_string is None:
            # If no pattern_string is present:
            err_s = 'The pattern_string needs to be configured'
            log.error(err_s)
            raise ValueError('The pattern_string needs to be configured')

    def exit(self):
        log.info('Exit: regex filter')

    def invoke(self, packet):
        if packet.data is None:
            return packet

        m = self.regex_object.match(packet.data)
        if m is not None:
            packet.data = m.groupdict()
        else:
            packet.data = {}

        return packet



import os
import sys

from stetl.etl import ETL
from stetl.outputs.standardoutput import StandardOutput
from stetl.splitter import Splitter
from tests.stetl_test_case import StetlTestCase

class SplitterMultiOutputTest(StetlTestCase):
    """Unit tests for Splitter"""
    pass

    def setUp(self):
        super(SplitterMultiOutputTest, self).setUp()

        # Initialize Stetl
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(curr_dir, 'configs/splittermultioutput.cfg')}
        self.etl = ETL(cfg_dict)

    def test_instance(self):
        chain = StetlTestCase.get_chain(self.etl)

        splitter_comp = chain.first_comp.next
        self.assertTrue(isinstance(splitter_comp, Splitter))

        # The next is a list of multiple Outputs
        self.assertEqual(len(splitter_comp.next), 2)
        self.assertTrue(isinstance(splitter_comp.next[0], StandardOutput))
        self.assertTrue(isinstance(splitter_comp.next[1], StandardOutput))

    def test_execute(self):
        chain = StetlTestCase.get_chain(self.etl)
        chain.run()

        # Result should be merged lines from both files
        result = sys.stdout.getvalue().split('\n')

        # Strip empty lines
        result = [s for s in result if (s or len(s) > 0)]

        # Total should be twice of linecount non-empty lines in input file
        self.assertEqual(len(result), 36)



# Main ETL program.
#
# Author: Just van den Broecke
#
import os
import re
import sys
from configparser import ConfigParser, ExtendedInterpolation
from io import StringIO
from . import version
from .util import Util
from .chain import Chain

log = Util.get_log('ETL')


class ETL:
    """The main class: builds ETL Chains with connected Components from a config and let them run.

    Usually this class is called via :mod:`main`  but it may be called directly for direct integration.

    """

    CONFIG_DIR = None

    def __init__(self, options_dict, args_dict=None):
        """
        :param options_dict: dictionary with options, now only config_file files with path to config file
        :param args_dict: optional dictionary with arguments to be substituted for symbolic values in config
        :return:

        Assume path to config .ini file is in options dict
        """
        # args_dict is optional and is used to do string substitutions in options_dict.config file

        log.info("INIT - Stetl version is %s" % str(version.__version__))

        self.options_dict = options_dict
        config_file = self.options_dict.get('config_file')

        if config_file is None or not os.path.isfile(config_file):
            log.error('No config file found at: %s' % config_file)
            sys.exit(1)

        ETL.CONFIG_DIR = os.path.dirname(os.path.abspath(config_file))
        log.info("Config/working dir = %s" % ETL.CONFIG_DIR)

        self.configdict = ConfigParser(interpolation=ExtendedInterpolation())

        sys.path.append(ETL.CONFIG_DIR)

        config_str = ''
        try:
            # Get config file as string
            log.info("Reading config_file = %s" % config_file)
            f = open(config_file, 'r')
            config_str = f.read()
            f.close()
        except Exception as e:
            log.error("Cannot read config file: err=%s" % str(e))
            raise e

        args_names = list()
        try:
            # Optional: expand symbolic arguments from args_dict and or OS Env
            # ignore errors here as { .. } may appear at random.

            # Parse unique list of argument names from config file string.
            # https://www.machinelearningplus.com/python/python-regex-tutorial-examples/
            args_names = list(set(re.findall(r'{[A-Z|a-z]\w+}', config_str)))
            args_names = [name.split('{')[1].split('}')[0] for name in args_names]

            # Optional: expand from equivalent env vars
            args_dict = self.env_expand_args_dict(args_dict, args_names)

            # In general all arg names should be present in args dict
            for args_name in args_names:
                if args_name not in args_dict:
                    log.warn("Arg not found in args nor environment: name=%s" % args_name)
                    # raise Exception("name=%s" % args_name)

        except Exception as e:
            log.warn("Expanding config arguments (non fatal yet): %s" % str(e))

        try:
            if args_dict:
                log.info("Substituting %d args in config file from args_dict: %s" % (len(args_names), str(args_names)))

                # Do replacements  see http://docs.python.org/2/library/string.html#formatstrings
                # and render substituted config string
                config_str = config_str.format(**args_dict)

                log.info("Substituting args OK")

        except Exception as e:
            log.error("Error substituting config arguments: err=%s" % str(e))
            raise e

        try:
            # Put Config string into buffer (readfp() needs a readline() method)
            config_buf = StringIO(config_str)

            # Parse config from file buffer
            self.configdict.read_file(config_buf, config_file)
        except Exception as e:
            log.error("Error populating config dict from config string: err=%s" % str(e))
            raise e

    def env_expand_args_dict(self, args_dict, args_names):
        """
        Expand values in dict with equivalent values from the
        OS Env. NB vars in OS Env should be prefixed with `STETL_` or `stetl_`
        as to get overrides by accident.

        :return: expanded args_dict or None
        """
        env_dict = os.environ
        for name in env_dict:
            args_key = '_'.join(name.split('_')[1:])
            if name.lower().startswith('stetl_') and args_key in args_names:
                # Get real key, e.g. "STETL_HOST" becomes "HOST"
                # "stetl_host" becomes "host".
                args_value = env_dict[name]
                if not args_dict:
                    args_dict = dict()

                # Set: optionally override any existing value
                args_dict[args_key] = args_value
                log.info("Set/override from env var: %s" % name)

        return args_dict

    def run(self):
        # The main ETL processing
        log.info("START")
        t1 = Util.start_timer("total ETL")

        # Get the ETL Chain pipeline config strings
        # Default is to use the section [etl], but may be overidden on cmd line

        config_section = self.options_dict.get('config_section')
        if config_section is None:
            config_section = 'etl'

        chains_str = self.configdict.get(config_section, 'chains')
        if not chains_str:
            raise ValueError('ETL chain entry not defined in section [etl]')

        # Multiple Chains may be specified in the config
        chains_str_arr = chains_str.split(',')
        for chain_str in chains_str_arr:
            # Build single Chain of components and let it run
            chain = Chain(chain_str, self.configdict)
            chain.assemble()

            # Run the ETL for this Chain
            chain.run()

        Util.end_timer(t1, "total ETL")

        log.info("ALL DONE")



import os
import re
import sys

from stetl.etl import ETL
from stetl.filters.xmlelementreader import XmlElementReader
from tests.stetl_test_case import StetlTestCase

class XmlElementReaderTest(StetlTestCase):
    """Unit tests for XmlElementReader"""

    def setUp(self):
        super(XmlElementReaderTest, self).setUp()

        # Initialize Stetl
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(curr_dir, 'configs/xmlelementreader.cfg')}
        self.etl = ETL(cfg_dict)
    
    def test_class(self):
        chain = StetlTestCase.get_chain(self.etl)
        section = StetlTestCase.get_section(chain, 1)
        class_name = self.etl.configdict.get(section, 'class')
        
        self.assertEqual('stetl.filters.xmlelementreader.XmlElementReader', class_name)
    
    def test_instance(self):
        chain = StetlTestCase.get_chain(self.etl)

        self.assertTrue(isinstance(chain.first_comp.next, XmlElementReader))
        
    def check_execution_result(self, chain, result, with_namespace):        
        section = StetlTestCase.get_section(chain, 1)
        element_tags = self.etl.configdict.get(section, 'element_tags')
        pattern = "<Element %s%s " % ('{[^}]+}' if with_namespace else '', element_tags)
        for elem in result:
            self.assertIsNotNone(re.match(pattern, elem))
        
    def test_execute(self):
        chain = StetlTestCase.get_chain(self.etl)
        chain.run()

        # Check the number of elements
        result = sys.stdout.getvalue().strip().split('\n')
        self.assertEqual(len(result), 3)
        
        # Check the actual elements
        self.check_execution_result(chain, result, True)
        
    def test_strip_namespaces(self):
        chain = StetlTestCase.get_chain(self.etl, 1)
        chain.run()
        
        # Check the number of elements
        result = sys.stdout.getvalue().strip().split('\n')
        self.assertEqual(len(result), 3)
        
        # Check the actual elements
        self.check_execution_result(chain, result, False)
        
    def test_no_namespace(self):
        chain = StetlTestCase.get_chain(self.etl, 2)
        chain.run()
        
        # Check the number of elements
        result = sys.stdout.getvalue().strip().split('\n')
        self.assertEqual(len(result), 3)
        
        # Check the actual elements
        self.check_execution_result(chain, result, False)



import os
import sys

from stetl.etl import ETL
from stetl.inputs.fileinput import VsiZipFileInput
from tests.stetl_test_case import StetlTestCase

class VsiZipFileInputTest(StetlTestCase):
    """Unit tests for VsiZipFileInput"""

    def setUp(self):
        super(VsiZipFileInputTest, self).setUp()

        # Initialize Stetl
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(curr_dir, 'configs/vsizipfileinput.cfg')}
        self.etl = ETL(cfg_dict)
    
    def test_class(self):
        chain = StetlTestCase.get_chain(self.etl)
        section = StetlTestCase.get_section(chain)
        class_name = self.etl.configdict.get(section, 'class')
        
        self.assertEqual('stetl.inputs.fileinput.VsiZipFileInput', class_name)
    
    def test_instance(self):
        chain = StetlTestCase.get_chain(self.etl)
        
        self.assertTrue(isinstance(chain.first_comp, VsiZipFileInput))
    
    def test_execute(self):
        chain = StetlTestCase.get_chain(self.etl)
        chain.run()
        
        result = sys.stdout.getvalue().strip().split('\n')
        self.assertEqual(len(result), 40)
                                    
    def test_zip_name_filter(self):
        chain = StetlTestCase.get_chain(self.etl, 1)
        chain.run()
        
        result = sys.stdout.getvalue().strip().split('\n')
        self.assertEqual(len(result), 18)

    def test_xml_name_filter(self):
        chain = StetlTestCase.get_chain(self.etl, 2)
        chain.run()

        result = sys.stdout.getvalue().strip().split('\n')
        self.assertEqual(len(result), 18)

    def test_xml_re_name_filter(self):
        chain = StetlTestCase.get_chain(self.etl, 3)
        chain.run()

        result = sys.stdout.getvalue().strip().split('\n')
        self.assertEqual(len(result), 2)

    def test_xml_single_name__nested_filter(self):
        chain = StetlTestCase.get_chain(self.etl, 4)
        chain.run()

        result = sys.stdout.getvalue().strip().split('\n')
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], '/vsizip/{tests/data/vsizipinput.zip}/Leveringsdocument-BAG-Extract.xml')

    def test_xml_single_name_nested_filter(self):
        chain = StetlTestCase.get_chain(self.etl, 5)
        chain.run()

        result = sys.stdout.getvalue().strip().split('\n')
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], '/vsizip/{/vsizip/{tests/data/vsizipinput.zip}/9999WPL15092020.zip}/0221WPL15092020-000001.xml')

    def test_read_xml_from_vsi_path(self):
        chain = StetlTestCase.get_chain(self.etl, 6)
        chain.run()

        result = sys.stdout.getvalue()
        self.assertEqual(len(result), 25152)
        # self.assertEqual(result[0], 'xml stuff')



import mock
import os
import shutil
import sys

from stetl.etl import ETL
from stetl.outputs.execoutput import Ogr2OgrExecOutput
from tests.stetl_test_case import StetlTestCase

class Ogr2OgrExecOutputTest(StetlTestCase):
    """Unit tests for Ogr2OgrExecOutput"""

    def setUp(self):
        super(Ogr2OgrExecOutputTest, self).setUp()

        # Initialize Stetl
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(curr_dir, 'configs/ogr2ogrexecoutput.cfg')}
        self.etl = ETL(cfg_dict)
    
    def test_class(self):
        chain = StetlTestCase.get_chain(self.etl)
        section = StetlTestCase.get_section(chain, -1)
        class_name = self.etl.configdict.get(section, 'class')
        
        self.assertEqual('stetl.outputs.execoutput.Ogr2OgrExecOutput', class_name)
    
    def test_instance(self):
        chain = StetlTestCase.get_chain(self.etl)

        self.assertTrue(isinstance(chain.cur_comp, Ogr2OgrExecOutput))
        
    def parse_command(self, command):
        """
        Parses the command line. With this regex the string is split at whitespace, except when
        whitespace is occurring between quotes. Because split keeps the capturing groups this way,
        every second item is removed from the list.
        """
        
        import re
        list = re.split(r"\s+(?=([^\"]*\"[^\"]*\")*[^\"]*$)", command)
        return list[0::2]
    
    @mock.patch('subprocess.call', autospec=True)
    def test_execute(self, mock_call):
        chain = StetlTestCase.get_chain(self.etl)
        chain.run()
        
        self.assertTrue(mock_call.called)
        self.assertEqual(1, mock_call.call_count)
        
        # Check command line
        args, kwargs = mock_call.call_args
        list = self.parse_command(args[0])
        self.assertEqual(len(list), 5)
        self.assertEqual(list[0], 'ogr2ogr')
        
        # Compare command line arguments with config
        section = StetlTestCase.get_section(chain, -1)
        file_path = self.etl.configdict.get(StetlTestCase.get_section(chain), 'file_path')
        
        # Certain options should not occur
        self.assertFalse('-spat' in list)
        self.assertFalse('-lco' in list)
        
        # Destination format
        self.assertTrue('-f' in list)
        f_idx = list.index('-f')
        dest_format = self.etl.configdict.get(section, 'dest_format')
        self.assertEqual(list[f_idx + 1], dest_format)
        
        # Destination datasource
        dest_data_source = self.etl.configdict.get(section, 'dest_data_source')
        self.assertEqual(list[f_idx + 2], dest_data_source)
        
        # Source datasource
        self.assertEqual(list[-1], file_path)
            
    @mock.patch('subprocess.call', autospec=True)
    def test_execute_lco(self, mock_call):
        chain = StetlTestCase.get_chain(self.etl, 1)
        chain.run()
        
        # Check command line
        args, kwargs = mock_call.call_args
        list = self.parse_command(args[0])
        self.assertEqual(len(list), 9)
        
        # Check layer creation options
        self.assertTrue('-lco' in list)
        lco_indices = [i for i, x in enumerate(list) if x == '-lco']
        self.assertEqual(len(lco_indices), 2)
        self.assertEqual(list[lco_indices[0] + 1], 'LAUNDER=YES')
        self.assertEqual(list[lco_indices[1] + 1], 'PRECISION=NO')
            
    @mock.patch('subprocess.call', autospec=True)
    def test_execute_extent(self, mock_call):
        chain = StetlTestCase.get_chain(self.etl, 2)
        chain.run()
        
        # Check command line
        args, kwargs = mock_call.call_args
        list = self.parse_command(args[0])
        self.assertEqual(len(list), 10)
        
        # Check spatial extent
        section = StetlTestCase.get_section(chain, -1)
        self.assertTrue('-spat' in list)
        spat_idx = list.index('-spat')
        spatial_extent = self.etl.configdict.get(section, 'spatial_extent')
        self.assertEqual(spatial_extent.split(), list[spat_idx + 1:spat_idx + 5])
            
    @mock.patch('subprocess.call', autospec=True)
    def test_execute_options(self, mock_call):
        chain = StetlTestCase.get_chain(self.etl, 3)
        chain.run()
        
        # Check command line
        args, kwargs = mock_call.call_args
        list = self.parse_command(args[0])
        self.assertEqual(len(list), 13)
        
        # Check spatial extent
        self.assertTrue('-append' in list)
        self.assertTrue('-gt' in list)
        self.assertTrue('--config' in list)
            
    @mock.patch('subprocess.call', autospec=True)
    def test_execute_gfs(self, mock_call):
        chain = StetlTestCase.get_chain(self.etl, 4)
        chain.run()
        
        # Check command line
        args, kwargs = mock_call.call_args
        list = self.parse_command(args[0])
        self.assertEqual(len(list), 5)
        
        # Check if GFS file exists, and clean it up
        file_path = self.etl.configdict.get(StetlTestCase.get_section(chain), 'file_path')
        file_ext = os.path.splitext(file_path) 
        gfs_path = file_ext[0] + '.gfs'
        self.assertTrue(os.path.exists(gfs_path))
        
        os.remove(gfs_path)
        self.assertFalse(os.path.exists(gfs_path))
        
    @mock.patch('subprocess.call', autospec=True)
    def test_execute_cleanup(self, mock_call):
        # Copy test file to temporary location, because it will be cleaned up
        orig_file_path = self.etl.configdict.get(StetlTestCase.get_section(StetlTestCase.get_chain(self.etl)), 'file_path')
        orig_file_ext = os.path.splitext(orig_file_path)
        temp_file_path = orig_file_ext[0] + "_temp" + orig_file_ext[1]
        shutil.copy(orig_file_path, temp_file_path)
        
        chain = StetlTestCase.get_chain(self.etl, 5)
        chain.run()
        
        # Check command line
        args, kwargs = mock_call.call_args
        list = self.parse_command(args[0])
        self.assertEqual(len(list), 5)
        
        # Check if temp file has been removed
        self.assertFalse(os.path.exists(temp_file_path))
        
    @mock.patch('subprocess.call', autospec=True)
    def test_execute_cleanup_gfs(self, mock_call):
        # Copy test file to temporary location, because it will be cleaned up
        orig_file_path = self.etl.configdict.get(StetlTestCase.get_section(StetlTestCase.get_chain(self.etl)), 'file_path')
        orig_file_ext = os.path.splitext(orig_file_path)
        temp_file_path = orig_file_ext[0] + "_temp" + orig_file_ext[1]
        shutil.copy(orig_file_path, temp_file_path)
        
        chain = StetlTestCase.get_chain(self.etl, 6)
        chain.run()
        
        # Check command line
        args, kwargs = mock_call.call_args
        list = self.parse_command(args[0])
        self.assertEqual(len(list), 5)
        
        # Check if temp file has been removed
        self.assertFalse(os.path.exists(temp_file_path))
        
        # Check if GFS file has already been removed
        gfs_path = orig_file_ext[0] + "_temp.gfs"
        self.assertFalse(os.path.exists(gfs_path))
        
    @mock.patch('subprocess.call', autospec=True)
    def test_execute_no_cleanup(self, mock_call):
        chain = StetlTestCase.get_chain(self.etl, 7)
        file_path = self.etl.configdict.get(StetlTestCase.get_section(chain), 'file_path')
        chain.run()
        
        # Check command line
        args, kwargs = mock_call.call_args
        list = self.parse_command(args[0])
        self.assertEqual(len(list), 5)
        
        # Check if input file still exists
        self.assertTrue(os.path.exists(file_path))



# -*- coding: utf-8 -*-
#
# Applies a GDAL/OGR virtual file system (VSI) filter to the input source.
# See for more information: https://www.gdal.org/gdal_virtual_file_systems.html
#
# Author: Frank Steggink
#
from stetl.filter import Filter
from stetl.util import Util
from stetl.packet import FORMAT

log = Util.get_log('vsifilter')


class VsiFilter(Filter):
    """
    Abstract base class for applying a GDAL/OGR virtual file system (VSI) filter.
    """

    # Start attribute config meta

    # End attribute config meta

    # Constructor
    def __init__(self, configdict, section, vsiname):
        Filter.__init__(self, configdict, section, consumes=FORMAT.record, produces=FORMAT.string)
        self.vsiname = vsiname

    def invoke(self, packet):

        if packet.data is None:
            log.info("No file name given")
            return packet

        file_path = packet.data['file_path'].replace('\\', '/')
        packet.data = '/%s/%s/%s' % (self.vsiname, file_path, packet.data['name'])
        return packet


class VsiZipFilter(VsiFilter):
    """
    Applies a VSIZIP filter to the input record.

    consumes=FORMAT.record, produces=FORMAT.string
    """

    # Start attribute config meta

    # End attribute config meta

    # Constructor
    def __init__(self, configdict, section):
        VsiFilter.__init__(self, configdict, section, 'vsizip')



# String filtering.
#
# Author:Just van den Broecke

from stetl.component import Config
from stetl.util import Util
from stetl.filter import Filter
from stetl.packet import FORMAT

log = Util.get_log("stringfilter")


class StringFilter(Filter):
    """
    Base class for any string filtering
    """

    # Constructor
    def __init__(self, configdict, section, consumes, produces):
        Filter.__init__(self, configdict, section, consumes, produces)

    def invoke(self, packet):
        if packet.data is None:
            return packet
        return self.filter_string(packet)

    def filter_string(self, packet):
        pass


class StringSubstitutionFilter(StringFilter):
    """
    String filtering using Python advanced String formatting.
    String should have substitutable values like {schema} {foo}
    format_args should be of the form format_args = schema:test foo:bar ...

    consumes=FORMAT.string, produces=FORMAT.string
    """

    @Config(ptype=str, default=None, required=True)
    def format_args(self):
        """
        Provides a list of format arguments used by the string substitution filter. Formatting of content according to Python String.format().
        String should have substitutable values like {schema} {foo}.

        Example: format_args = schema:test foo:bar
        """
        pass

    @Config(ptype=str, default=':', required=False)
    def separator(self):
        """
        Provides the separator to split the format argument names from their values.
        """
        pass

    # Constructor
    def __init__(self, configdict, section):
        StringFilter.__init__(self, configdict, section, consumes=FORMAT.string, produces=FORMAT.string)

        # Convert string to dict: http://stackoverflow.com/a/1248990
        self.format_args_dict = Util.string_to_dict(self.format_args, self.separator)

    def filter_string(self, packet):
        # String substitution based on Python String.format()
        packet.data = packet.data.format(**self.format_args_dict)
        return packet


class StringConcatFilter(StringFilter):
    """
    Concatenates a specified string with the input string (packet.data) either/or as prefix (prepend)
    or postfix (append) and outputs that concatenation.

    consumes=FORMAT.string, produces=FORMAT.string
    """

    @Config(ptype=str, default=None, required=False)
    def append_string(self):
        """
        String to be appended.

        Example: append_string = /002PND150904.xml
        """
        pass

    @Config(ptype=str, default=None, required=False)
    def prepend_string(self):
        """
        String to be prepended.

        Example: prepend_string = /vsizip/
        """
        pass

    # Constructor
    def __init__(self, configdict, section):
        StringFilter.__init__(self, configdict, section, consumes=FORMAT.string, produces=FORMAT.string)

    def filter_string(self, packet):
        if not packet.data:
            return

        if self.append_string:
            packet.data = packet.data + self.append_string

        if self.prepend_string:
            packet.data = self.prepend_string + packet.data

        return packet



#!/usr/bin/python

## Copyright (c) 2011 Astun Technology

## Permission is hereby granted, free of charge, to any person obtaining a copy
## of this software and associated documentation files (the "Software"), to deal
## in the Software without restriction, including without limitation the rights
## to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
## copies of the Software, and to permit persons to whom the Software is
## furnished to do so, subject to the following conditions:

## The above copyright notice and this permission notice shall be included in
## all copies or substantial portions of the Software.

## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
## IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
## FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
## AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
## LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
## OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
## THE SOFTWARE.

"""
A collection of classes used to manipulate Ordnance Survey GB GML data,
used with prepgml4ogr.py. (But here used/called from Stetl Filter)
"""

import os
from lxml import etree


class prep_osgml():
    """
    Base class that provides the main interface methods `prepare_feature` and
    `get_feat_types` and performs basic manipulation such as exposing the fid,
    adding and element containing the filename of the source and adding an
    element with the orientation in degrees.

    """
    def __init__(self, inputfile):
        self.inputfile = inputfile
        self.feat_types = []

    def get_feat_types(self):
        return self.feat_types

    def prepare_feature(self, feat_str):

        # Parse the xml string into something useful
        feat_elm = etree.fromstring(feat_str)
        feat_elm = self._prepare_feat_elm(feat_elm)

        return etree.tostring(feat_elm,
                              encoding='UTF-8',
                              pretty_print=True).decode('utf_8')

    def _prepare_feat_elm(self, feat_elm):

        feat_elm = self._add_fid_elm(feat_elm)
        feat_elm = self._add_filename_elm(feat_elm)
        feat_elm = self._add_orientation_degree_elms(feat_elm)

        return feat_elm

    def _add_fid_elm(self, feat_elm):

        # Create an element with the fid
        elm = etree.SubElement(feat_elm, "fid")
        elm.text = feat_elm.get('fid')

        return feat_elm

    def _add_filename_elm(self, feat_elm):

        # Create an element with the filename
        elm = etree.SubElement(feat_elm, "filename")
        elm.text = os.path.basename(self.inputfile)

        return feat_elm

    def _add_orientation_degree_elms(self, feat_elm):

        # Correct any orientation values to be a
        # tenth of their original value
        orientation_elms = feat_elm.xpath('//orientation')
        for elm in orientation_elms:
            # Add a new orientDeg element as a child to the
            # the orientation elm to be orientation/10
            # (this applies integer division which is fine in
            # this instance as we are not concerned with the decimals)
            degree_elm = etree.SubElement(elm.getparent(), "orientDeg")
            degree_elm.text = str(int(elm.text) / 10)

        return feat_elm


class prep_vml(prep_osgml):
    """
    Preperation class for OS VectorMap Local features.

    """
    def __init__(self, inputfile):
        prep_osgml.__init__(self, inputfile)
        self.feat_types = [
            'Text',
            'VectorMapPoint',
            'Line',
            'RoadCLine',
            'Area'
        ]


class prep_osmm_topo(prep_osgml):
    """
    Preperation class for OS MasterMap features which in addition to the work
    performed by `prep_osgml` adds `themes`, `descriptiveGroups` and
    `descriptiveTerms` elements containing a delimited string of the attributes
    that can appear multiple times.

    """
    def __init__(self, inputfile):
        prep_osgml.__init__(self, inputfile)
        self.feat_types = [
            'BoundaryLine',
            'CartographicSymbol',
            'CartographicText',
            'TopographicArea',
            'TopographicLine',
            'TopographicPoint'
        ]
        self.list_seperator = ', '

    def _prepare_feat_elm(self, feat_elm):

        feat_elm = prep_osgml._prepare_feat_elm(self, feat_elm)
        feat_elm = self._add_lists_elms(feat_elm)

        return feat_elm

    def _add_lists_elms(self, feat_elm):

        feat_elm = self._create_list_of_terms(feat_elm, 'theme')
        feat_elm = self._create_list_of_terms(feat_elm, 'descriptiveGroup')
        feat_elm = self._create_list_of_terms(feat_elm, 'descriptiveTerm')

        return feat_elm

    def _create_list_of_terms(self, feat_elm, name):
        text_list = feat_elm.xpath('//%s/text()' % name)
        if len(text_list):
            elm = etree.SubElement(feat_elm, "%ss" % name)
            elm.text = self.list_seperator.join(text_list)
        return feat_elm


class prep_osmm_topo_qgis(prep_osmm_topo):
    """
    Preperation class for OS MasterMap features which in addition to the work performed by
    `prep_osmm_topo` adds QGIS specific label attributes such as `qFont` and `aAnchorPos`.

    """

    def __init__(self, filename):
        prep_osmm_topo.__init__(self, filename)

        # AC - define the font
        if os.name is 'posix':
            # Will probably need different font names
            self.fonts = ('Garamond', 'Arial', 'Roman', 'ScriptC')
        elif os.name is 'nt':
            # Ordnance Survey use
            #   'Lutheran', 'Normal', 'Light Roman', 'Suppressed text'
            self.fonts = ('GothicE', 'Monospac821 BT', 'Consolas', 'ScriptC', 'Arial Narrow')
        elif os.name is 'mac':
            # Will probably need different font name
            self.fonts = ('Garamond', 'Arial', 'Roman', 'ScriptC')

        # AC - the possible text placement positions used by QGIS
        self.anchorPosition = ('Bottom Left', 'Left', 'Top Left', 'Bottom',
                                'Over', 'Top',  'Bottom Right', 'Right', 'Top Right')

    def _prepare_feat_elm(self, feat_elm):

        feat_elm = prep_osmm_topo._prepare_feat_elm(self, feat_elm)
        feat_elm = self._add_qgis_elms(feat_elm)

        return feat_elm

    def _add_qgis_elms(self, feat_elm):

        if feat_elm.tag == 'CartographicText':
            text_render_elm = feat_elm.xpath('//textRendering')[0]

            anchor_pos = int(text_render_elm.xpath('./anchorPosition/text()')[0])
            try:
                anchor_pos = self.anchorPosition[anchor_pos]
            except:
                anchor_pos = 4
            elm = etree.SubElement(text_render_elm, 'qAnchorPos')
            elm.text = anchor_pos

            font = int(text_render_elm.xpath('./font/text()')[0])
            try:
                font = self.fonts[font]
            except:
                font = 'unknown font (%s)' % str(font)
            elm = etree.SubElement(text_render_elm, 'qFont')
            elm.text = font

        return feat_elm


class prep_osmm_itn(prep_osgml):
    """
    Preperation class for OS MasterMap ITN features.

    """

    def __init__(self, filename):

        prep_osgml.__init__(self, filename)

        self.feat_types = [
            'FerryLink',
            'FerryNode',
            'InformationPoint',
            'Road',
            'RoadNode',
            'RoadNodeInformation',
            'RoadLink',
            'RoadLinkInformation',
            'RoadRouteInformation'
        ]

    def _prepare_feat_elm(self, feat_elm):

        feat_elm = prep_osgml._prepare_feat_elm(self, feat_elm)
        feat_elm = self._expose_links(feat_elm)

        return feat_elm

    def _expose_links(self, feat_elm):

        link_list = feat_elm.xpath('//networkMember | //directedLink | //directedNode | //referenceToRoadLink | //referenceToRoadNode | //referenceToTopographicArea')
        for elm in link_list:
            for name in elm.attrib:
                value = elm.get(name)
                if name == 'href':
                    name = '%s_%s' % (elm.tag, 'ref')
                    value = value[1:]
                elif name == 'orientation':
                    name = '%s_%s' % (elm.tag, name)
                    value = '1' if value == '+' else '0'
                sub_elm = etree.SubElement(elm, name)
                sub_elm.text = value

        return feat_elm


class prep_addressbase():
    """
    Simple preperation of AddressBase data

    """
    def __init__(self, inputfile):
        self.inputfile = inputfile
        self.feat_types = ['Address']

    def get_feat_types(self):
        return self.feat_types

    def prepare_feature(self, feat_str):

        # Parse the xml string into something useful
        feat_elm = etree.fromstring(feat_str)
        feat_elm = self._prepare_feat_elm(feat_elm)

        return etree.tostring(feat_elm,
                              encoding='UTF-8',
                              pretty_print=True).decode('utf_8')

    def _prepare_feat_elm(self, feat_elm):

        feat_elm = self._drop_gmlid(feat_elm)

        return feat_elm

    def _drop_gmlid(self, feat_elm):

        feat_elm.attrib.pop('id')

        return feat_elm



# Output classes for ETL, executing commands.
#
# Author: Frank Steggink
#
import subprocess
import os
import re
import shutil
from stetl.component import Config
from stetl.output import Output
from stetl.util import Util
from stetl.packet import FORMAT

log = Util.get_log('execoutput')


class ExecOutput(Output):
    """
    Executes any command (abstract base class).
    """

    @Config(ptype=str, default='', required=False)
    def env_args(self):
        """
        Provides of list of environment variables which will be used when executing the given command.

        Example: env_args = pgpassword=postgres othersetting=value~with~spaces
        """
        pass

    @Config(ptype=str, default='=', required=False)
    def env_separator(self):
        """
        Provides the separator to split the environment variable names from their values.
        """
        pass

    def __init__(self, configdict, section, consumes):
        Output.__init__(self, configdict, section, consumes)

    def write(self, packet):
        return packet

    def execute_cmd(self, cmd):
        env_vars = Util.string_to_dict(self.env_args, self.env_separator)
        old_environ = os.environ.copy()

        try:
            os.environ.update(env_vars)
            log.info("executing cmd=%s" % Util.safe_string_value(cmd))
            subprocess.call(cmd, shell=True)
            log.info("execute done")
        finally:
            os.environ = old_environ


class CommandExecOutput(ExecOutput):
    """
    Executes an arbitrary command.

    consumes=FORMAT.string
    """

    def __init__(self, configdict, section):
        ExecOutput.__init__(self, configdict, section, consumes=FORMAT.string)

    def write(self, packet):
        if packet.data is not None:
            self.execute_cmd(packet.data)

        return packet


class Ogr2OgrExecOutput(ExecOutput):
    """
    Executes an Ogr2Ogr command.
    Input is a file name to be processed.
    Output by calling Ogr2Ogr command.

    consumes=FORMAT.string
    """

    # Start attribute config meta
    # Applying Decorator pattern with the Config class to provide
    # read-only config values from the configured properties.

    @Config(ptype=str, default=None, required=True)
    def dest_data_source(self):
        """
        String denoting the OGR data destination. Usually a path to a file like "path/rivers.shp" or connection string
        to PostgreSQL like "PG: host=localhost dbname='rivers' user='postgres'".
        """
        pass

    @Config(ptype=str, default=None, required=False)
    def dest_format(self):
        """
        Instructs GDAL to use driver by that name to open data destination. Not required for
        many standard formats that are self-describing like ESRI Shapefile.

        Examples: 'PostgreSQL', 'GeoJSON' etc
        """
        pass

    @Config(ptype=str, default=None, required=False)
    def lco(self):
        """
        Options for newly created layer (-lco).
        """

        pass

    @Config(ptype=str, default=None, required=False)
    def spatial_extent(self):
        """
        Spatial extent (-spat), to pass as xmin ymin xmax ymax
        """
        pass

    @Config(ptype=str, default=None, required=False)
    def gfs_template(self):
        """
        Name of GFS template file to use during loading. Passed to ogr2ogr as
        --config GML_GFS_TEMPLATE <name>
        """
        pass

    @Config(ptype=str, default=None, required=False)
    def options(self):
        """
        Miscellaneous options to pass to ogr2ogr.
        """
        pass

    @Config(ptype=bool, default=False, required=False)
    def cleanup_input(self):
        """
        Flag to indicate whether the input file to ogr2ogr should be cleaned up.
        """
        pass

    @Config(ptype=bool, default=False, required=False)
    def always_apply_lco(self):
        """
        Flag to indicate whether the layer creation options should be applied
        to all runs.
        """
        pass

    # End attribute config meta

    def __init__(self, configdict, section):
        ExecOutput.__init__(self, configdict, section, consumes=[FORMAT.string, FORMAT.gdal_vsi_path])
        self.regex_vsi_filter = re.compile("^/vsi[a-z0-9_]+/.*", re.I)

        self.ogr2ogr_cmd = 'ogr2ogr -f ' + self.dest_format + ' ' + self.dest_data_source

        if self.spatial_extent:
            self.ogr2ogr_cmd += ' -spat ' + self.spatial_extent
        if self.options:
            self.ogr2ogr_cmd += ' ' + self.options

        self.first_run = True

    def write(self, packet):
        if packet.data is None:
            return packet

        # Execute ogr2ogr
        ogr2ogr_cmd = self.ogr2ogr_cmd
        if self.lco and (self.first_run or self.always_apply_lco):
            ogr2ogr_cmd += ' ' + self.lco
            self.first_run = False

        if type(packet.data) is list:
            for item in packet.data:
                self.execute(ogr2ogr_cmd, item)
        else:
            self.execute(ogr2ogr_cmd, packet.data)

        return packet

    def execute(self, ogr2ogr_cmd, file_path):

        # For creating tables the GFS file needs to be newer than
        # the .gml file. -lco GML_GFS_TEMPLATE somehow does not work
        # so we copy the .gfs file each time with the .gml file with
        # the same base name
        # Copy the .gfs file if required, use the same base name
        # so ogr2ogr will pick it up.
        # Always assemble the GFS path, in case it is provided from outside.
        # Note that for now using a GFS file is not supported with a VSI filter.
        file_ext = os.path.splitext(file_path)
        gfs_path = file_ext[0] + '.gfs'
        if self.gfs_template and not self.regex_vsi_filter.match(file_path):
            shutil.copy(self.gfs_template, gfs_path)

        # Append file name to command as last argument
        self.execute_cmd(ogr2ogr_cmd + ' ' + file_path)

        if self.cleanup_input and not self.regex_vsi_filter.match(file_path):
            os.remove(file_path)
            if gfs_path and os.path.exists(gfs_path):
                os.remove(gfs_path)



import mock
import os
import shutil
import sys

from stetl.etl import ETL
from tests.stetl_test_case import StetlTestCase

class VsiFilterTest(StetlTestCase):
    """Unit tests for VsiFilter"""

    def setUp(self):
        super(VsiFilterTest, self).setUp()

        # Initialize Stetl
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(curr_dir, 'configs/vsifiltertest.cfg')}
        self.etl = ETL(cfg_dict)

    def parse_command(self, command):
        """
        Parses the command line. With this regex the string is split at whitespace, except when
        whitespace is occurring between quotes. Because split keeps the capturing groups this way,
        every second item is removed from the list.
        """

        import re
        list = re.split(r"\s+(?=([^\"]*\"[^\"]*\")*[^\"]*$)", command)
        return list[0::2]

    @mock.patch('subprocess.call', autospec=True)
    def test_vsizipfilter(self, mock_call):
        chain = StetlTestCase.get_chain(self.etl, 1)
        chain.run()

        # Check command line
        args, kwargs = mock_call.call_args
        list = self.parse_command(args[0])

        # Check whether the vsizip filter has been applied to the input file
        self.assertTrue(list[-1].startswith('/vsizip/tests/data/bestuurlijkegrenzen.zip/'))

    @mock.patch('subprocess.call', autospec=True)
    def test_vsizipfilter_cleanup_input(self, mock_call):
        """
        This test checks whether Stetl won't show an error when it tries to clean up the input file,
        which it can't find. This is the full path to the file within the ZIP file prepended with
        /vsizip/.
        """

        chain = StetlTestCase.get_chain(self.etl, 2)
        chain.run()

        # Check command line
        args, kwargs = mock_call.call_args
        list = self.parse_command(args[0])

        # Check whether the vsizip filter has been applied to the input file
        self.assertTrue(list[-1].startswith('/vsizip/tests/data/bestuurlijkegrenzen.zip/'))



# Input classes for ETL.
#
# Author: Just van den Broecke
#
from .util import Util
from .component import Component

log = Util.get_log('input')


class Input(Component):
    """
    Abstract Base class for all Input Components.

    """

    def __init__(self, configdict, section, produces):
        Component.__init__(self, configdict, section, consumes=None, produces=produces)

        log.info("cfg = %s" % self.cfg.to_string())

    def invoke(self, packet):
        return self.read(packet)

    def read(self, packet):
        return packet



#!/usr/bin/env python
#
# Output classes for ETL.
#
# Author: Just van den Broecke
#
import subprocess
import os
import re
import shutil

from stetl.component import Config
from stetl.output import Output
from stetl.util import Util, gdal, ogr, osr
from stetl.packet import FORMAT

log = Util.get_log('ogroutput')


class OgrOutput(Output):
    """
    Direct GDAL OGR output via Python OGR wrapper. Via the Python API http://gdal.org/python
    OGR Features are written.

    This output can write almost any geospatial, OGR-defined, dataformat.

    consumes=FORMAT.ogr_feature or FORMAT.ogr_feature_array
    """

    # Start attribute config meta
    # Applying Decorator pattern with the Config class to provide
    # read-only config values from the configured properties.

    @Config(ptype=bool, default=False, required=False)
    def append(self):
        """
        Add to destination destination if it extists (ogr2ogr -append option).
        """

        pass

    @Config(ptype=str, default=None, required=True)
    def dest_data_source(self):
        """
        String denoting the OGR data destination. Usually a path to a file like "path/rivers.shp" or connection string
        to PostgreSQL like "PG: host=localhost dbname='rivers' user='postgres'".
        """
        pass

    @Config(ptype=str, default=None, required=False)
    def dest_format(self):
        """
        Instructs GDAL to use driver by that name to open data destination. Not required for
        many standard formats that are self-describing like ESRI Shapefile.

        Examples: 'PostgreSQL', 'GeoJSON' etc
        """
        pass

    @Config(ptype=list, default=[], required=False)
    def dest_create_options(self):
        """
        Creation options.

        Examples: ..
        """
        pass

    @Config(ptype=dict, default=None, required=False)
    def dest_options(self):
        """
        Custom data destination-specific options. Used in gdal.SetConfigOption().
        """
        pass

    @Config(ptype=list, default=[], required=False)
    def layer_create_options(self):
        """
        Options for newly created layer (-lco).
        """

        pass

    @Config(ptype=str, default=None, required=True)
    def new_layer_name(self):
        """
        Layer name for layer created in the destination source.

        """

        pass

    @Config(ptype=bool, default=False, required=False)
    def overwrite(self):
        """
        Overwrite destination if it extists (ogr2ogr -overwrite option).
        """

        pass

    @Config(ptype=str, default=None, required=False)
    def target_srs(self):
        """
        SRS (projection) for the target.
        """
        pass

    @Config(ptype=str, default=None, required=False)
    def sql(self):
        """
        String with SQL query. Mandatory for PostgreSQL OGR dest.
        """
        pass

    @Config(ptype=bool, default=False, required=False)
    def always_apply_lco(self):
        """
        Flag to indicate whether the layer creation options should be applied
        to all runs.
        """
        pass

    # End attribute config meta

    # Constructor
    def __init__(self, configdict, section):
        Output.__init__(self, configdict, section, consumes=[FORMAT.ogr_feature, FORMAT.ogr_feature_array])

    def init(self):

        self.ogr = ogr
        # http://trac.osgeo.org/gdal/wiki/PythonGotchas
        self.gdal = gdal
        self.gdal.UseExceptions()
        log.info("Using GDAL/OGR version: %d" % int(gdal.VersionInfo('VERSION_NUM')))

        # GDAL error handler function
        # http://pcjericks.github.io/py-gdalogr-cookbook/gdal_general.html
        def gdal_error_handler(err_class, err_num, err_msg):
            err_type = {
                gdal.CE_None: 'None',
                gdal.CE_Debug: 'Debug',
                gdal.CE_Warning: 'Warning',
                gdal.CE_Failure: 'Failure',
                gdal.CE_Fatal: 'Fatal'
            }
            err_msg = err_msg.replace('\n', ' ')
            err_class = err_type.get(err_class, 'None')
            log.error('Error Number: %s, Type: %s, Msg: %s' % (err_num, err_class, err_msg))

        # install error handler
        self.gdal.PushErrorHandler(gdal_error_handler)

        # Raise a dummy error for testing
        # self.gdal.Error(1, 2, 'test error')

        self.update = self.overwrite or self.append

        if self.dest_options:
            for k in self.dest_options:
                self.gdal.SetConfigOption(k, self.dest_options[k])

        self.dest_driver = None
        self.dest_fd = None

        # Loosely based on https://github.com/OSGeo/gdal/blob/trunk/gdal/swig/python/samples/ogr2ogr.py

        # /* -------------------------------------------------------------------- */
        # /*      Try opening the output data source as an existing, writable      */
        # /* -------------------------------------------------------------------- */
        if self.update:
            # Try opening in update mode
            self.dest_fd = ogr.Open(self.dest_data_source, True)

            if self.dest_fd is not None:
                if len(self.dest_create_options) > 0:
                    log.warn("Datasource creation options ignored since an existing datasource being updated.")

                self.dest_driver = self.dest_fd.GetDriver()
                if self.overwrite:
                    self.dest_driver.DeleteDataSource(self.dest_data_source)
                    self.dest_fd = None
                    self.dest_driver = None
                    self.update = False

        # /* -------------------------------------------------------------------- */
        # /*      Find the output driver.                                         */
        # /* -------------------------------------------------------------------- */
        if self.dest_driver is None:

            # Open OGR data dest in write-only mode.
            self.dest_driver = ogr.GetDriverByName(self.dest_format)

            # Report failure if failed
            if self.dest_driver is None:
                log.error("Cannot open OGR data destination: %s with the following drivers." % self.dest_data_source)

                for iDriver in range(self.ogr.GetDriverCount()):
                    log.info("  ->  " + self.ogr.GetDriver(iDriver).GetName())

                raise Exception()

            if self.dest_driver.TestCapability(ogr.ODrCCreateDataSource) is False:
                log.error("%s driver does not support data source creation." % self.dest_format)
                raise Exception()

        # /* -------------------------------------------------------------------- */
        # /*      Create the output data source.                                  */
        # /* -------------------------------------------------------------------- */
        if self.dest_fd is None:
            self.dest_fd = self.dest_driver.CreateDataSource(self.dest_data_source, options=self.dest_create_options)
            if self.dest_fd is None:
                log.error("%s driver failed to create %s" % (self.dest_format, Util.safe_string_value(self.dest_data_source)))
                raise Exception()

        # /* -------------------------------------------------------------------- */
        # /*      Parse the output SRS definition if possible.                    */
        # /* -------------------------------------------------------------------- */
        output_srs_ref = None
        if self.target_srs is not None:
            output_srs_ref = osr.SpatialReference()
            if output_srs_ref.SetFromUserInput(self.target_srs) != 0:
                log.error("Failed to process SRS definition: %s" % self.target_srs)
            raise Exception()

        self.layer = self.dest_fd.CreateLayer(self.new_layer_name, output_srs_ref, ogr.wkbUnknown,
                                              self.layer_create_options)
        self.feature_def = None

        log.info("Opened OGR dest ok: %s " % Util.safe_string_value(self.dest_data_source))

    def write(self, packet):

        # Are we all done?
        if packet.data is None or self.dest_fd is None:
            self.write_end(packet)
            return packet

        if self.layer is None:
            log.info("No Layer, end writing to: %s" % Util.safe_string_value(self.dest_data_source))
            return packet

        # Assume ogr_feature_array input, otherwise convert ogr_feature to list
        if type(packet.data) is list:
            # Write feature collection to OGR Layer output
            for feature in packet.data:
                self.write_feature(feature)

            self.write_end(packet)

        else:
            # Write single feature to OGR Layer output
            if packet.end_of_stream or packet.end_of_doc:
                self.write_end(packet)
                return packet

            self.write_feature(packet.data)

        return packet

    def write_feature(self, feature):

        # Set the Layer Feature Definition once, using the first feature received
        if self.feature_def is None:
            self.feature_def = feature.GetDefnRef()
            field_count = self.feature_def.GetFieldCount()
            for i in range(0, field_count):
                field_def = self.feature_def.GetFieldDefn(i)
                self.layer.CreateField(field_def)

        # Write Feature to the Layer
        self.layer.CreateFeature(feature)

        # Dispose memory
        feature.Destroy()

    def write_end(self, packet):
        # Destroy not required anymore: http://trac.osgeo.org/gdal/wiki/PythonGotchas
        # self.dest_fd.Destroy()
        log.info("End writing to: %s" % Util.safe_string_value(self.dest_data_source))
        self.dest_fd = None
        self.layer = None
        return packet


class Ogr2OgrOutput(Output):
    """
    Output from GML etree doc to any OGR2OGR output using the GDAL/OGR ogr2ogr command

    consumes=FORMAT.etree_doc
    """

    def __init__(self, configdict, section):
        Output.__init__(self, configdict, section, consumes=FORMAT.etree_doc)

        self.temp_file = self.cfg.get('temp_file')
        self.regex_vsi_filter = re.compile("^/vsi[a-z0-9_]+/.*", re.I)

        # For creating tables the GFS file needs to be newer than
        # the .gml file. -lco GML_GFS_TEMPLATE somehow does not work
        # so we copy the .gfs file each time with the .gml file with
        # the same base name
        self.gfs_file = self.cfg.get('gfs_file')
        self.lco = self.cfg.get('lco')
        self.spatial_extent = self.cfg.get('spatial_extent')
        self.ogr2ogr_cmd = self.cfg.get('ogr2ogr_cmd').replace('\\\n', ' ').replace('\n', ' ')
        if self.spatial_extent:
            self.ogr2ogr_cmd += ' -spat ' + self.spatial_extent
        self.first_run = True

    def save_doc(self, packet, file_path):
        if packet.data is None:
            return packet

        log.info('writing to file %s' % file_path)
        out_file = open(file_path, 'w')
        out_file.writelines(packet.to_string())
        out_file.close()

        # Copy the .gfs file if required, use the same base name
        # so ogr2ogr will pick it up.
        # Note that for now using a GFS file is not supported with a VSI filter.
        if self.gfs_file and not self.regex_vsi_filter.match(file_path):
            file_ext = os.path.splitext(file_path)
            shutil.copy(self.gfs_file, file_ext[0] + '.gfs')

        log.info("written to %s" % file_path)
        return packet

    def execute_cmd(self, cmd):
        use_shell = True
        if os.name == 'nt':
            use_shell = False

        log.info("executing cmd=%s" % cmd)
        subprocess.call(cmd, shell=use_shell)
        log.info("execute done")

    def write(self, packet):
        if packet.data is None:
            return packet

        # Save the doc to a temp file
        self.save_doc(packet, self.temp_file)

        # Execute ogr2ogr
        ogr2ogr_cmd = self.ogr2ogr_cmd
        if self.lco and (self.first_run or self.always_apply_lco):
            ogr2ogr_cmd += ' ' + self.lco
            self.first_run = False

        self.execute_cmd(ogr2ogr_cmd)
        return packet



#!/usr/bin/env python
#
# Filter: XML validation.
#
# NB: you need to have installed libxml2 2.8.0 or newer!
# Older libxml2 versions like 2.7.8 have a bug which causes failure in GML Schema
# parsing. See https://bugzilla.gnome.org/show_bug.cgi?id=630130
#
# Author:Just van den Broecke
#
from stetl.util import Util, etree
from stetl.filter import Filter
from stetl.packet import FORMAT

log = Util.get_log("xmlvalidator")


class XmlSchemaValidator(Filter):
    """
    Validates an etree doc and prints result to log.

    consumes=FORMAT.etree_doc, produces=FORMAT.etree_doc
    """

    # Constructor
    def __init__(self, configdict, section):
        Filter.__init__(self, configdict, section, consumes=FORMAT.etree_doc, produces=FORMAT.etree_doc)
        self.enabled = self.cfg.get_bool('enabled', True)
        self.xsd = self.cfg.get('xsd')
        log.info("Building the Schema once with (GML XSD) dependencies for schema=%s (be patient...)" % self.xsd)
        self.schema = etree.XMLSchema(etree.parse(self.xsd))

    def invoke(self, packet):
        if packet.data is None or not self.enabled:
            return packet
        return self.validate(packet)

    def validate(self, packet):
        log.info("Validating doc against schema=%s ..." % self.xsd)
        result = self.schema.validate(packet.data)
        log.info("Validation result: %s" % str(result))
        return packet



# testing: to be called by nosetests

import os

from stetl.etl import ETL
from stetl.chain import Chain
from tests.stetl_test_case import StetlTestCase

class ChainTest(StetlTestCase):
    """Basic chain tests"""

    def setUp(self):
        super(ChainTest, self).setUp()

        # Initialize Stetl
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(curr_dir, 'configs/copy_in_out.cfg')}
        self.etl = ETL(cfg_dict)
        
    def test_chain(self):
        chain = StetlTestCase.get_chain(self.etl, assemble=False)
        
        self.assertIsNone(chain.first_comp)
        self.assertIsNone(chain.cur_comp)

    def test_chain_assembly(self):
        chain = StetlTestCase.get_chain(self.etl, assemble=False)
        chain.assemble()
        
        self.assertIsNotNone(chain.first_comp)
        self.assertIsNotNone(chain.cur_comp)

        self.assertIsNotNone(chain.first_comp.next)
        self.assertIsNone(chain.cur_comp.next)
        
        comp = chain.first_comp
        while comp.next is not None:
            comp = comp.next
            
        self.assertIs(comp, chain.cur_comp)



import os
import sys

from stetl.etl import ETL
from stetl.packet import Packet
from stetl.inputs.fileinput import GlobFileInput
from tests.stetl_test_case import StetlTestCase

class GlobFileInputTest(StetlTestCase):
    """Unit tests for GlobFileInput"""

    def setUp(self):
        super(GlobFileInputTest, self).setUp()

        # Initialize Stetl
        curr_dir = os.path.dirname(os.path.realpath(__file__))
        cfg_dict = {'config_file': os.path.join(curr_dir, 'configs/globfileinput.cfg')}
        self.etl = ETL(cfg_dict)
    
    def test_class(self):
        chain = StetlTestCase.get_chain(self.etl)
        section = StetlTestCase.get_section(chain)
        class_name = self.etl.configdict.get(section, 'class')
        
        self.assertEqual('stetl.inputs.fileinput.GlobFileInput', class_name)
    
    def test_instance(self):
        chain = StetlTestCase.get_chain(self.etl)
        
        self.assertTrue(isinstance(chain.first_comp, GlobFileInput))
    
    def test_execute(self):
        chain = StetlTestCase.get_chain(self.etl)
        chain.run()
        
        result = sys.stdout.getvalue().strip().split('\n')
        self.assertEqual(len(result), 1)
        
    def test_wildcard(self):
        chain = StetlTestCase.get_chain(self.etl, 1)
        chain.run()
        
        result = sys.stdout.getvalue().strip().split('\n')
        self.assertEqual(len(result), 2)



# Expands an archive file into a collection of files.
#
# Author: Just van den Broecke 2021
#
import os.path
from stetl.component import Config
from stetl.filter import Filter
from stetl.util import Util
from stetl.packet import FORMAT

log = Util.get_log('archiveexpander')


class ArchiveExpander(Filter):
    """
    Abstract Base Class.
    Expands an archive file into a collection of files.

    consumes=FORMAT.string, produces=FORMAT.string
    """

    # Start attribute config meta

    @Config(ptype=str, default='temp_dir', required=True)
    def target_dir(self):
        """
        Target directory to write the extracted files to.
        """
        pass

    @Config(ptype=bool, default=False, required=False)
    def remove_input_file(self):
        """
        Delete input archive file when the chain has been completed?
        """
        pass

    @Config(ptype=bool, default=True, required=False)
    def clear_target_dir(self):
        """
        Delete the files from the target directory  when the chain has been completed?
        """
        pass

    # End attribute config meta

    # Constructor
    def __init__(self, configdict, section, consumes, produces):
        Filter.__init__(self, configdict, section, consumes=consumes, produces=produces)
        self.input_archive_file = None
        if not os.path.exists(self.target_dir):
            os.mkdir(self.target_dir)

    def remove_file(self, file_path):
        if os.path.isfile(file_path):
            os.remove(file_path)

    def wipe_dir(self, dir_path):
        if os.path.isdir(dir_path):
            for file_object in os.listdir(dir_path):
                file_object_path = os.path.join(dir_path, file_object)
                if os.path.isdir(file_object_path):
                    self.wipe_dir(file_object_path)
                    os.rmdir(file_object_path)
                    return

                os.remove(file_object_path)

    def expand_archive(self, packet):
        log.error('Only classes derived from ArchiveExpander can be used!')

    def invoke(self, packet):

        if packet.data is None:
            log.info("Input data is empty")
            return packet

        # Optionally clear target dir
        self.wipe_dir(self.target_dir)

        self.input_archive_file = packet.data

        # Let derived class provide archive expansion (.zip, .tar etc)
        self.expand_archive(self.input_archive_file)
        if not os.listdir(self.target_dir):
            log.warn('No expanded files in {}'.format(self.target_dir))
            packet.data = None
            return packet

        # ASSERT: expanded files in target dir
        file_count = len(os.listdir(self.target_dir))
        log.info('Expanded {} into {} OK - filecount={}'.format(
            self.input_archive_file, self.target_dir, file_count))

        # Output the target dir path where expanded files are found
        packet.data = self.target_dir

        return packet

    def after_chain_invoke(self, packet):
        if self.remove_input_file:
            self.remove_file(self.input_archive_file)

        if self.clear_target_dir:
            self.wipe_dir(self.target_dir)

        return True


class ZipArchiveExpander(ArchiveExpander):
    """
    Extracts all files from a ZIP file into the configured  target directory.

    consumes=FORMAT.string, produces=FORMAT.string
    """

    def __init__(self, configdict, section):
        ArchiveExpander.__init__(self, configdict, section, consumes=FORMAT.string, produces=FORMAT.string)

    def expand_archive(self, file_path):

        import zipfile
        if not file_path.lower().endswith('zip'):
            log.warn('No zipfile passed: {}'.format(file_path))
            return

        zipfile.ZipFile(file_path).extractall(path=self.target_dir)



import logging
import os
import sys
import unittest

from stetl.chain import Chain
from stetl.util import Util
from io import StringIO

logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)

class StetlTestCase(unittest.TestCase):
    """Base class for Stetl test cases"""

    def setUp(self):        
        # Replace logger method
        self._old_get_log = Util.get_log
        
        @staticmethod
        def get_log_new(name, level=logging.WARN):
            log = logging.getLogger(name)
            log.setLevel(level)
            return log
            
        Util.get_log = get_log_new
    
        # Disable info logging
        logging.disable(logging.WARN)
        
        # Replace stdout
        self._saved_stdout = sys.stdout
        sys.stdout = StringIO()
        
        # Replace work dir
        self._cwd = os.getcwd()
        os.chdir(os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))
        
    def tearDown(self):
        # Restore old enviroment
        sys.stdout = self._saved_stdout
        logging.disable(logging.NOTSET)
        Util.get_log = self._old_get_log
        os.chdir(self._cwd)

    @classmethod
    def get_chain(cls, etl, index=0, assemble=True):
    
        chains_str = etl.configdict.get('etl', 'chains')
        chain_strs = chains_str.split(',')        
        chain = Chain(chain_strs[index].strip(), etl.configdict)
        
        if assemble:
            chain.assemble()
        
        return chain
        
    @classmethod
    def get_section(cls, chain, index=0):
        return chain.chain_str.split('|')[index]



# Reads an XML file and returns XML elements.
# Based on inputs.fileinput.XmlElementStreamFileInput.
#
# Author: Frank Steggink
#
from copy import deepcopy

from stetl.component import Config
from stetl.filter import Filter
from stetl.util import Util, etree
from stetl.packet import FORMAT

log = Util.get_log('xmlelementreader')


class XmlElementReader(Filter):
    """
    Extracts XML elements from a file, outputs each feature element in Packet.
    Parsing is streaming (no internal DOM buildup) so any file size can be handled.
    Use this class for your big GML files!

    consumes=FORMAT.string, produces=FORMAT.etree_element
    """

    # Start attribute config meta
    @Config(ptype=list, default=None, required=True)
    def element_tags(self):
        """
        Comma-separated string of XML (feature) element tag names of the elements that should be extracted
        and added to the output element stream.
        """
        pass

    @Config(ptype=bool, default=False, required=False)
    def strip_namespaces(self):
        """
        should namespaces be removed from the input document and thus not be present in the output element stream?
        """
        pass

    # End attribute config meta

    # Constructor
    def __init__(self, configdict, section):
        Filter.__init__(self, configdict, section, consumes=FORMAT.string, produces=FORMAT.etree_element)
        self.context = None
        self.root = None
        self.cur_file_path = None
        self.elem_count = 0
        log.info("Element tags to be matched: %s" % self.element_tags)

    def invoke(self, packet):
        if packet.data is None:
            log.info("No XML file given")
            return packet

        if self.cur_file_path is None:
            self.cur_file_path = packet.data

        event = None
        packet.data = None

        if self.context is None:
            # Open file
            fd = open(self.cur_file_path, 'rb')
            self.elem_count = 0
            log.info("file opened : %s" % self.cur_file_path)
            self.context = etree.iterparse(fd, events=("start", "end"))
            self.context = iter(self.context)
            event, self.root = next(self.context)

        packet = self.process_xml(packet)

        return packet

    def process_xml(self, packet):
        while self.context is not None:
            # while not packet.is_end_of_doc():
            try:
                event, elem = next(self.context)
            except (etree.XMLSyntaxError, StopIteration):
                # workaround for etree.XMLSyntaxError https://bugs.launchpad.net/lxml/+bug/1185701
                self.context = None

            if self.context is None:
                # Always end of doc
                # TODO: is this still useful for a non-input component?
                packet.set_end_of_doc()
                log.info("End of doc: %s elem_count=%d" % (self.cur_file_path, self.elem_count))

                return packet

            # Filter out Namespace from the tag
            # this is the easiest way to go for now
            tag = elem.tag.split('}')
            if len(tag) == 2:
                # Namespaced tag: 2nd is tag
                tag = tag[1]
            else:
                # Non-namespaced tag: first
                tag = tag[0]

            if tag in self.element_tags:
                if event == "start":
                    pass
                elif event == "end":
                    packet.data = deepcopy(elem)
                    self.elem_count += 1

                    if self.strip_namespaces:
                        packet.data = Util.stripNamespaces(elem).getroot()

                    # Clear the element which has been read. Don't clear the root document,
                    # since the last element hasn't been processed yet.
                    elem.clear()

            # If there is a next component, let it process
            if self.next:
                # Hand-over data (line, doc whatever) to the next component
                packet.format = self._output_format
                packet = self.next.process(packet)

        return packet
